"""
File Purpose: reading moments for EppicCalculator.
"""
import re

import xarray as xr

from .eppic_io_tools import read_moments_out_files
from ...dimensions import INPUT_SNAP, CHARGED
from ...errors import LoadingNotImplementedError, ValueLoadingError
from ...quantities import QuantityLoader
from ...tools import simple_property, UNSET, xarray_sum
from ...defaults import DEFAULTS


class EppicMomentsLoader(QuantityLoader):
    '''moments (from moments*.out files) from Eppic outputs.'''

    # # # MOMENTS FROM FILE # # #
    @known_var(dims=['snap', 'fluid', 'component'])
    def get_moment1(self):
        '''1st moment of distribution function, along an axis (vector component). Mean.
        directly from moments*.out files (multiplied by units conversion factor).
        Compare with 'nmean_u' (which is equivalent to 'mean_(n*u)/mean_n').
            (nmean instead of mean, because moments.out averages across all particles!
            u is the value averaged across all particles in each cell,
            and mean_u averages u across all cells, using an equal weight for each cell.
            so mean_u gives less weight to particles in denser cells, unlike nmean_u and moments.out.)
        '''
        result = self.load_across_dims(self._load_moment, 'moment1', dims=['fluid', 'component'])
        return self.record_units(result * self.u('speed')**1)

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_moment2(self):
        '''2nd moment of distribution function, along an axis (vector component). Variance.
        directly from moments*.out files (multiplied by units conversion factor).
        Compare with 'nmean_(vsqr-(nmean_u)**2)'.
            (nmean instead of mean, because moments.out averages across all particles!
            vsqr is the value averaged across all particles in each cell,
            and mean_vsqr averages vsqr across all cells, using an equal weight for each cell.
            So, mean_vsqr gives less weight to particles in denser cells, unlike nmean_vsqr and moments.out.)
        '''
        result = self.load_across_dims(self._load_moment, 'moment2', dims=['fluid', 'component'])
        return self.record_units(result * self.u('speed')**2)

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_moment3(self):
        '''3rd moment of distribution function, along an axis (vector component).
        directly from moments*.out files (multiplied by units conversion factor).
        '''
        result = self.load_across_dims(self._load_moment, 'moment3', dims=['fluid', 'component'])
        return self.record_units(result * self.u('speed')**3)

    @known_var(dims=['snap', 'fluid', 'component'])
    def get_moment4(self):
        '''4th moment of distribution function, along an axis (vector component).
        directly from moments*.out files (multiplied by units conversion factor).
        '''
        result = self.load_across_dims(self._load_moment, 'moment4', dims=['fluid', 'component'])
        return self.record_units(result * self.u('speed')**4)

    input_moments = simple_property('_input_moments', setdefaultvia='_read_all_moments',
        doc=r'''moments from domain000/moments*.out files.
        Can be read using PlasmaCalcs architecture by reading vars:
            moment1, moment2, moment3, moment4.
        These are sensitive to self.component and self.fluid.
        To get moments values directly from file, without helpful xarray labeling,
            use the "eppic_var" versions, which look like:
            'moment{i}{x}{N}' (e.g. 'moment3x0'), where
                i=1,2,3,4 is the moment number,
                x='x','y','z' is the component,
                N=0,1,2,... is the fluid number (i.e., distribution number).
        That is similar to getting, e.g., 'fluxx1'; it gives a value directly from file.

        [EFF] for efficiency, input moments are only read from file one time, then cached.
            You can remove the cached values by doing: del self.input_moments.''')

    def _read_all_moments(self):
        '''read all moments; return result. See help(type(self).input_moments) for more details.'''
        return read_moments_out_files(self.input_deck.dirname)

    def _loadable_moments(self):
        '''tells eppic_var variable names of loadable moments.'''
        moments = self.input_moments
        dists = list(moments.keys())
        # note - the next line assumes all 4*3 moments exist in each file:
        return [f'moment{i}{x}{N}' for i in (1,2,3,4) for N in dists for x in 'xyz']

    def _load_moment(self, eppic_var, *args__None, snap=UNSET, **kw__None):
        r'''load moment indicated by eppic_var. (e.g. 'moment3x0' --> moment 3, component x, fluid 0).
        Units will be the same units as the moments.out files (raw units).
        
        eppic_var: str
            should match full_pattern='moment([1234])([xyz])(\d+)', or partial_pattern='moment([1234])'
            matches full_pattern -->
                ([1234]) indices moment number (1,2,3,4),
                ([xyz]) indicates component (x, y, z),
                (\d+) indicates distribution number (any integer).
            matches partial_pattern --> 
                try to match full_pattern using eppic_var=self._var_for_load_fromfile(eppic_var) instead.
                (_var_for_load_fromfile appends x and N, if self._loading_component & self._loading_fluid)
            match not found -->
                raise LoadingNotImplementedError
            match found but no associated value in moments data -->
                raise ValueLoadingError
        snap: UNSET, None, or any indicator of any number of snaps
            if provided, return value at this snap, else return value at self.snap.
            if snap indicates multiple values, result will be a list of values.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        full_pattern = r'moment([1234])([xyz])(\d+)'  # e.g. 'moment3x0'
        match = re.fullmatch(full_pattern, eppic_var)
        if match is None:
            partial_match = re.fullmatch(r'moment([1234])', eppic_var)
            if partial_match:
                var_dimmed = self._var_for_load_fromfile(eppic_var)
                match = re.fullmatch(full_pattern, var_dimmed)
            if match is None:
                raise LoadingNotImplementedError(f'var={eppic_var!r} not recognized by _load_moment')
        else:
            var_dimmed = eppic_var
        momi, x, N = match.groups()

        # [TODO] if moments is refactored to be more object-oriented, the code here can be cleaned up a bit.
        moments = self.input_moments
        momname = {'1': 'Mean', '2': 'Variance', '3': 'Moment3', '4': 'Moment4'}[momi]
        momkey = f'V{x}_{momname}'
        try:
            momvals = moments[N][momkey]
        except KeyError:
            raise ValueLoadingError(f'{var_dimmed!r} not found in moments data') from None
        kw_snap = dict() if snap is UNSET else dict(snap=snap)
        with self.using(**kw_snap):
            if self.snap_is_iterable():
                result = [(momvals[int(s.file_s(self))] if s.exists_for(self) else self.snap_dim.NAN) for s in self.snap]
                result = xr.DataArray(result, dims=['snap'], attrs=dict(units='raw'))  # moments file uses raw units
                result = self.assign_snap_along(result, 'snap')
            else:
                result = momvals[int(self.snap.file_s(self))] if self.snap.exists_for(self) else self.snap_dim.NAN
                result = xr.DataArray(result, attrs=dict(units='raw'))  # moments file uses raw units
                result = self.assign_snap_coord(result)
        return result

    # # # TEMPERATURES # # #
    @known_pattern(r'Ta(joule)?_from_moment2', deps=['m', 'moment2'])
    def get_Ta_from_moment2_or_Tajoule_from_moment2(self, var, *, _match=None):
        '''temperature ("anisotropic"), from moment2 instead of nvsqr/n.
        'Ta_from_moment2'      --> (m * moment2 / kB) [Kelvin]
        'Tajoule_from_moment2' --> (m * moment2)      [energy units]
        Compare with nmean_Ta (or nmean_Tajoule).
            nmean instead of mean because moments.out averages across all particles!
            (see help(self.get_moment2) for more details.)
        '''
        here, = _match.groups()
        result = self('m') * self('moment2')
        if here != 'joule':  # Ta, not Tajoule
            result = result / self.u('kB')
        return result

    @known_pattern(r'T(joule)?_from_moment2', deps=[{0: "Ta{group0}_from_moment2"}])
    def get_T_from_moment2_or_Tjoule_from_moment2(self, var, *, _match=None):
        '''temperature ("isotropic/maxwellian"), from moment2 instead of nvsqr/n.
        'T_from_moment2' --> mean of Ta_from_moment2 across components.
            == (m * moment2 / kB) [Kelvin], averaged across components.
        'Tjoule_from_moment2' --> similar, but don't divide by kB; result has [energy units].
        '''
        with self.using(component=None):
            return self('Ta_from_moment2').mean('component')

    @known_var(deps=['rmscomps_Ta_from_moment2'], ignores_dims=['component'])
    def get_T_box(self):
        '''temperature of the entire simulation box, as if full of particles,
        and observed by something that could not resolve the individual cells.
        Equivalent: rmscomps(Ta_from_moment2).

        Ignores Ta components for directions in which the box has no extent.
        '''
        return self('rmscomps_Ta_from_moment2', component=self.maindims)

    # # # "ZEROTH ORDER" NEUTRAL HEATING # # #
    @known_var(deps=['m_n', 'nuns', 'mod2_u_drift'], ignores_dims=['snap'])
    def get_dTndt0_s(self):
        '''zeroth order rate of heating of neutrals due to collisions with s.
        Purely based on the input deck values.
        dTndt0_s = (2 m_n / (3 kB)) * nuns * |u0_s|^2,
            where u0_s is the zeroth order drift velocities (see u_drift)
            |u0_s|^2 = (kappa_s^2 / (1 + kappa_s^2)) * EBspeed^2,
                where EBspeed = E_perpmod_B / |B|.

        Does the calculation at the INPUT_SNAP. Removes 'snap' dim from result.
        (Future implementation note: might want to allow computing at other snaps, using means)

        Result should be equivalent to self('frominputs_dTndt_s'), aside from rounding errors.
        '''
        with self.using(snap=INPUT_SNAP):
            coeff0 = (2 * self('m_n') / (3 * self.u('kB')))
            result = coeff0 * self('nuns') * self('mod2_u_drift')
        return result.drop_vars('snap')

    @known_var(deps=['dTndt0_s'], ignores_dims=['fluid'])
    def get_dTndt0(self):
        '''zeroth order rate of heating of neutrals due to collisions with charged species.
        dTndt0 = sum_s dTndt0_s == sum_s (2 m_n / (3 kB)) * nuns * |u0_s|^2.
        See dTndt0_s for more details.

        Result should be equivalent to self('frominputs_dTndt'), aside from rounding errors.
        '''
        return xarray_sum(self('dTndt0_s', fluid=CHARGED), 'fluid')

    # # # OTHER VALUES DERIVED FROM MOMENTS # # #
    @known_var(deps=['moment1', 'n0', 'q'])
    def get_Jf_from_moment1(self):
        '''self.fluid's contribution to current density, based on moment1 info: q * n0 * moment1.
        (compare to self('Jf'), which tells q * n * u.)
        '''
        return self('q') * self('n0') * self('moment1')

    @known_var(deps=['sum_fluids_Jf_from_moment1'])
    def get_J_from_moment1(self):
        '''total current density, based on moment1 info: sum_fluids_Jf_from_moment1.
        Always sums across all charged fluids from self.fluids.
        (compare to self('J'), which tells sum_fluids(q * n * u).)
        '''
        return self('sum_fluids_Jf_from_moment1', fluid=CHARGED)



"""
File Purpose: KNOWN_PLOTTERS for EppicInstabilityCalculator
"""

import os

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from ....errors import PlottingAmbiguityError
from ....plotting import PlotterManager, plot_note
from ....tools import (
    simple_property,
    code_snapshot_info, Stopwatch,
    xr1d,
)


class EppicInstabilityPlotterManager(PlotterManager):
    '''EppicInstabilityCalculator plotting routines'''
    @known_plotter(kinds=['timelines', 'n_mul', 'runtime_info'], cost=10)
    def plot_min_n_nodes(ec, **kw_plot_settings):
        '''timelines of min_n_nodes vs n_mul (log2-log10 plot).
        Includes lines for:
            min_n_nodes_given_nsubdomains
            min_n_nodes_given_runtime_guess
        '''
        defaults_here = dict(robust=False, markersize=8)
        plt.figure(figsize=(6,3))
        arr = ec('log2_min_n_nodes_given_nsubdomains')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='(given nsubdomains)',
                         marker='o', fillstyle='none',
                         **{**defaults_here, **kw_plot_settings})
        plt.grid()
        # given runtime guess
        arr = ec('log2_min_n_nodes_given_runtime_guess')
        arr.pc.timelines(tCOORD, label='(given runtime guess)',
                         marker='x',
                         **{**defaults_here, **kw_plot_settings})
        # better labels
        plt.title(f'min_n_nodes\n{ec.title}', fontsize='small')
        pow2s = list(range(2,9))
        plt.yticks(pow2s, labels=[2**p for p in pow2s])
        plt.ylabel('min_n_nodes')
        plt.xlabel('log10_n_mul')
        return result

    @known_plotter(kinds=['timelines', 'n_mul', 'runtime_info'], cost=15)
    def plot_safe_node_hours(ec, **kw_plot_settings):
        '''timelines of safe_node_hours vs n_mul (log-log plot).
        Includes line for n_nodes = min, 4, 16, 64, 256.

        Compare with SU cost on Frontera (1 node-hour is 1 SU in 'normal' queue).
        '''
        defaults_here = dict(robust=False)
        plt.figure(figsize=(6,3))
        min_n_nodes = ec('min_n_nodes')
        arr = ec('log10_safe_node_hours', n_nodes=min_n_nodes)
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='n_nodes=(min_n_nodes)',
                         marker='X', color='black', ls='--', fillstyle='none',
                         **{**defaults_here, 'markersize': 8, **kw_plot_settings})
        plt.grid()
        # for a variety of n_nodes
        n_nodes = xr1d([4, 16, 64, 256], 'n_nodes')
        arr = ec('log10_safe_node_hours', n_nodes=n_nodes)
        arr.pc.timelines(tCOORD, marker='o', ls='-', alpha=0.5,
                         **{**defaults_here, 'markersize': 4, **kw_plot_settings})
        # better labels
        plt.title(f'log10_safe_node_hours\n{ec.title}', fontsize='small')
        plt.xlabel('log10_n_mul')
        # note about cost / cost(4). (i.e. "how much more expensive is this run than with n_nodes=4?)
        valrat = (10**(arr - arr.isel(n_nodes=0))).pc.squeeze_close(tol=0.05)
        if valrat.ndim==1:  # only varies in the n_nodes dim
            n0 = arr.coords['n_nodes'][0].item()
            strs = [f'{val.coords["n_nodes"]:3d} | {val.item():.2g}' for val in valrat]
            plot_note(f'nodes | cost÷cost({n0}):\n  '+'\n  '.join(strs),
                         xy='outside lower right', font='monospace', fontsize='small')
        return result

    @known_plotter(kinds=['timelines', 'n_mul', 'runtime_info'], cost=17)
    def plot_safe_runtime_seconds(ec, **kw_plot_settings):
        '''timelines of safe_runtime_seconds vs n_mul (log-log plot).
        Includes line for n_nodes = min, 4, 16, 64, 256.
        Also includes horizontal lines at 1 hour, 8 hours, and 48 hours.

        Compare with actual wall-clock time it takes to run the run.
        '''
        defaults_here = dict(robust=False)
        plt.figure(figsize=(6,3))
        # hlines (style: do 48 first since it will be above the others on the y-axis too)
        plt.axhline(np.log10(48 * 3600), color='gray', ls='--', label='(48 hours)')
        plt.axhline(np.log10(8 * 3600), color='gray', ls='-.', label='(8 hours)')
        plt.axhline(np.log10(1 * 3600), color='gray', ls=':', label='(1 hour)')
        # n_nodes=min
        min_n_nodes = ec('min_n_nodes')
        arr = ec('log10_safe_runtime_seconds', n_nodes=min_n_nodes)
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='n_nodes=(min_n_nodes)',
                         marker='X', color='black', ls='--', fillstyle='none',
                         **{**defaults_here, 'markersize': 8, **kw_plot_settings})
        plt.grid()
        # for a variety of n_nodes
        n_nodes = xr1d([4, 16, 64, 256], 'n_nodes')
        arr = ec('log10_safe_runtime_seconds', n_nodes=n_nodes)
        arr.pc.timelines(tCOORD, marker='o', ls='-', alpha=0.5,
                         **{**defaults_here, 'markersize': 4, **kw_plot_settings})
        # better labels
        plt.title(f'log10_safe_runtime_seconds\n{ec.title}', fontsize='small')
        plt.xlabel('log10_n_mul')
        plt.legend(bbox_to_anchor=(1.02, 0.95), loc='upper left', fontsize='small')
        # note about time(4) / time. (i.e. "how much less time does this run take than with n_nodes=4?)
        valrat = (10**(arr.isel(n_nodes=0) - arr)).pc.squeeze_close(tol=0.05)
        if valrat.ndim==1:
            n0 = arr.coords['n_nodes'][0].item()
            strs = [f'{val.coords["n_nodes"]:3d} | {val.item():.2g}' for val in valrat]
            plot_note(f'nodes | time({n0})÷time:\n  '+'\n  '.join(strs),
                         xy='outside lower right', font='monospace', fontsize='x-small')
        return result

    @known_plotter(kinds=['timelines', 'n_mul', 'runtime_info'], cost=11)
    def plot_safe_total_Gbytes(ec, **kw_plot_settings):
        '''timeline of safe_total_Gbytes vs n_mul (log-log plot).'''
        plt.figure(figsize=(5,2.5))
        arr = ec('log10_safe_total_Gbytes')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, marker='o', color='black')
        plt.grid()
        # better labels
        plt.title(f'log10_safe_total_Gbytes\n{ec.title}', fontsize='small')
        plt.xlabel('log10_n_mul')
        return result



import math
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams
from matplotlib.animation import FuncAnimation, FFMpegWriter

import PlasmaCalcs

# SET PATH TO FFMPEG
working_from_home = False

laptop_ffmpeg_path = r'C:\\Users\\ratta\\PycharmProjects\\pythonProject\\1D_PIC\\EM1D_PIC\\ffmpeg\\bin\\ffmpeg.exe'
bu_ffmpeg_path = r'C:\\Users\\ratta\\PycharmProjects\\EM1D\\ffmpeg\\bin\\ffmpeg.exe'

if working_from_home:
    ffmpeg_path = laptop_ffmpeg_path
else:
    ffmpeg_path = bu_ffmpeg_path

# SET VIDEO WRITER PATH
rcParams['animation.ffmpeg_path'] = ffmpeg_path

# INPUT PARAMETERS
path_to_folder = r"G:\My Drive\Research\eppic_out\eregion"
#folder_names = {"FBI, normal": "FBI_normal"}
folder_names = {"FBI, normal": "FBI_normal", "FBI, Maxwell": "FBI_maxwell", "laminar, normal": "laminar_normal",
                "laminar, Maxwell": "laminar_maxwell"}  # key: nickname, value: folder name
n_columns = 4
species_to_plot = {1: "Ion"}  # key: number, value: name
plot_quantity = "u"
plot_every = 80
n_bins = 64
plot_range = [[-2000, 2000], [-2000, 2000]] #  [[xmin, xmax], [ymin, ymax]]
slices = (("x", "y"), ("y", "z"), ("x", "z"))
mp4_name = f"test{''.join(map(str, species_to_plot))}"


# FUNCTION TO MERGE RIGHT SUBPLOTS (modified from a function in Save2541/1DEMPyPIC/animation.py)
def merge_right_subplots(fig, axs):
    """
    Merge all subplots in the right column to create a tall subplot for the colorbar
    :param fig: figure
    :param axs: axes
    :return: merged right axis
    """
    gs = axs[0, -1].get_gridspec()
    # Remove the underlying axes
    for ax in axs[:, -1]:
        ax.remove()
    return fig.add_subplot(gs[:, -1])


# CREATE CANVAS
n_plots = len(folder_names) * len(species_to_plot) * len(slices)
n_rows = math.ceil(n_plots / n_columns)
column_ratios = [1] * n_columns
column_ratios.append(0.05)
fig, axs = plt.subplots(nrows=n_rows, ncols=n_columns + 1, width_ratios=column_ratios, squeeze=False, figsize=[20, 11.25])
cb_ax = merge_right_subplots(fig, axs)


# LOAD INPUT DECKS
ec_list = []
snap_list = []
for run in folder_names:
    this_path = os.path.join(path_to_folder, folder_names[run], "eppic.i")
    ec = PlasmaCalcs.EppicCalculator.from_here(this_path, kw_units={"M": 1}, u_l=1)
    ec_list.append(ec)
    sample_snaps = ec.snaps[::plot_every]
    snap_list.append(sample_snaps)

# GET NUMBER OF TIME STEPS TO RUN (EQUAL TO THE TIME STEP OF THE RUN WITH THE LEAST TIME STEPS)
number_of_time_steps_to_run = min([len(x) for x in snap_list])

# FUNCTION TO GET THE NEXT ROW, COLUMN
def get_next_coordinates(current_coordinates=None, shape=(n_rows, n_columns), reset=False):
    """
    Get next (row, column) from the current (row, current).
    """
    if reset:
        return 0, 0
    current_row, current_column = current_coordinates
    number_of_rows, number_of_columns = shape
    if current_row < number_of_rows - 1:
        current_row += 1
    elif current_column < number_of_columns - 1:
        current_column += 1
        current_row = 0
    else:
        return 0, 0
    return current_row, current_column


# GET PLOT TITLES AND AXIS LABELS
plot_titles = {}
x_labels = {}
y_labels = {}
this_row, this_column = get_next_coordinates(reset=True)
for description in folder_names:
    for species in species_to_plot:
        species_name = species_to_plot[species]
        for components in slices:
            x_labels.setdefault(this_row, {})[this_column] = f"v{components[0]} (m/s)"
            y_labels.setdefault(this_row, {})[this_column] = f"v{components[1]} (m/s)"
            plot_titles.setdefault(this_row, {})[this_column] = f"{species_name} velocity distribution ({description})"
            this_row, this_column = get_next_coordinates((this_row, this_column))


# CONTAINER OF PLOT OBJECTS
plot_list = []
title_list = []
x_edges_list = []
y_edges_list = []
cb_list = []

# INITIALIZE PLOTS
def initialize_plots():
    """
    Initialize plots
    """
    current_coordinates = get_next_coordinates(reset=True)
    data_list = []
    time_list = []
    for run_number in range(len(ec_list)):
        ec = ec_list[run_number]
        this_snap = snap_list[run_number][0]
        ec.snap = this_snap
        for species in species_to_plot:
            ec.fluid = species
            for components in slices:
                data_pair = []
                for component in components:
                    ec.component = component
                    # GET DATA
                    array = ec(plot_quantity).values.ravel()
                    data_pair.append(array)
                data_list.append(data_pair)
                time_list.append(this_snap.t)

    for plot_number in range(len(time_list)):
        # GET AXIS
        this_axis = axs[current_coordinates[0], current_coordinates[1]]
        # GET PLOT TITLE
        title = plot_titles[current_coordinates[0]][current_coordinates[1]] + f" (Time = {time_list[plot_number]} s.)"
        # PLOT ON AXIS
        xdata, ydata = data_list[plot_number]
        _, x_edges, y_edges, plot = this_axis.hist2d(xdata, ydata, bins=n_bins, range=plot_range, norm="log", density=True)
        this_axis.set_xlabel(x_labels[current_coordinates[0]][current_coordinates[1]])
        this_axis.set_ylabel(y_labels[current_coordinates[0]][current_coordinates[1]])
        # SET TITLE & ADD TO TITLE LIST
        title_list.append(this_axis.set_title(title))
        # ADD TO PLOT LIST
        plot_list.append(plot)
        # ADD TO EDGES LIST
        x_edges_list.append(x_edges)
        y_edges_list.append(y_edges)
        # UPDATE ROW & COLUMN
        current_coordinates = get_next_coordinates(current_coordinates)

    # TIGHT LAYOUT
    plt.tight_layout()

    # CREATE COLOR BAR
    cb_list.append(plt.colorbar(plot_list[0], cax=cb_ax))


# EXTRACT DATA (FOR EACH TIME STEP)
def animate(time_step):
    """
    Extract data for each time step
    """
    current_coordinates = get_next_coordinates(reset=True)
    data_list = []
    time_list = []
    for run_number in range(len(ec_list)):
        ec = ec_list[run_number]
        this_snap = snap_list[run_number][time_step]
        ec.snap = this_snap
        for species in species_to_plot:
            ec.fluid = species
            for components in slices:
                data_pair = []
                for component in components:
                    ec.component = component
                    # GET DATA
                    try:
                        array = ec(plot_quantity).values.ravel()
                    except:
                        old_snap = snap_list[run_number][time_step - 1]
                        ec.snap = old_snap
                        array = ec(plot_quantity).values.ravel()
                        ec.snap = this_snap
                    data_pair.append(array)
                data_list.append(data_pair)
                time_list.append(f"{this_snap.t:.4f}")

    for plot_number in range(len(time_list)):
        # UPDATE PLOT
        this_plot = plot_list[plot_number]
        data_pair = data_list[plot_number]
        data = np.histogram2d(*data_pair, bins=[x_edges_list[plot_number], y_edges_list[plot_number]])[0]
        this_plot.set_array(data)
        # UPDATE TITLE
        this_title = title_list[plot_number]
        title = plot_titles[current_coordinates[0]][current_coordinates[1]] + f" (Time = {time_list[plot_number]} s.)"
        this_title.set_text(title)
        # UPDATE ROW & COLUMN
        current_coordinates = get_next_coordinates(current_coordinates)

    # UPDATE COLOR BAR
    #cb_list[0].mapable.set_clim(minimum, maximum)

    # TIGHT LAYOUT
    #plt.tight_layout()

    # PRINT TIME STEP
    print(time_step)

    # RETURN ARTISTS
    return plot_list, title_list, cb_list


# ANIMATE
anim = FuncAnimation(fig, animate, init_func=initialize_plots, save_count=number_of_time_steps_to_run)
# SAVE ANIMATION
f = os.path.join(path_to_folder, f"{mp4_name}.mp4")
os.makedirs(os.path.dirname(f), exist_ok=True)
writer_video = FFMpegWriter(fps=24)
anim.save(f, writer=writer_video)
print(f"Finished writing to {f}.")
plt.close()



"""
Package purpose: misc tools; portable (not specific to PlasmaCalcs).
"""
from .array_select import (
    select_i_closest, select_closest,
    select_i_before, select_before,
    select_i_after, select_after,
    select_i_between, select_between,
    ArraySelectable, ArraySelectableChildHaver,
)
from .arrays import (
    memory_size_check, memory_size_check_loading_arrays_like,
    finite_op, finite_min, finite_mean, finite_max, finite_std,
    finite_median, finite_percentile,
    unique_close,
    interprets_fractional_indexing, ndindex_array,
    looks_flat, nest_shape,
    wraplist, ndenumerate_nonNone,
    np_dtype_object_to_str,
)
from .display import (
    repr_simple,
    print_clear, help_str,
    join_strs_with_max_line_len,
)
from .docs_tools import (
    format_docstring,
    docstring_but_strip_examples,
    DocstringInfo,
    sphinx_docstring,
    list_objs,
)
from .history import (
    git_hash_local, git_hash, git_hash_here, git_hash_PlasmaCalcs,
    _PlasmaCalcs_version,
    datetime_now,
    code_snapshot_info,
)
from .fft_tools import (
    # array_fft
    fftN, fft2, fft1, fftfreq_shifted,
    ifftN, ifftfreq_shifted,
    # fft_dimnames
    FFTDimname,
    # fft_slices
    FFTSlices,
    # xarray_fft
    xarray_fftN,
    xarray_ifftN,
    xarray_lowpass,
)
from .imports import (
    enable_reload, reload,
    import_relative,
    ImportFailed,
)
from .io_tools import (
    attempt_literal_eval,
    read_idl_params_file, updated_idl_params_file,
    read_python_params_file,
)
from .iterables import (
    is_iterable,
    argmax, rargmax,
    scalar_item,
    DictWithAliases,
    DictlikeFromKeysAndGetitem,
    Partition, PartitionFromXarray,
    Container, ContainerOfList, ContainerOfArray, ContainerOfDict,
    Bijection, BijectiveMemory,
    SymmetricPairMapping,
    DictOfSimilar,
)
from .math import (
    ast_math_eval,
    round_to_int, float_rounding,
    is_integer,
    product, nonempty_product,
    np_all_int,
    as_roman_numeral, from_roman_numeral,
)
from .multiprocessing import (
    Task,
    CrashIfCalled, UniqueTask, UNSET_TASK, identity, IdentityTask,
    TaskContainer, TaskList, TaskArray,
    TaskContainerCallKwargsAttrHaver,
    TaskGroup, TaskPartition,
    mptest_add100, mptest_sleep, mptest_sleep_add100, mptest_echo,
    check_pickle, copy_via_pickle,
)
from .oop_tools import (
    # binding
    bind_to, Binding,
    # manage_attrs
    maintaining_attrs, using_attrs,
    maintain_attrs, use_attrs,
    MaintainingAttrs, UsingAttrs,
    UsingAttrsSignaled,
    IncrementableAttrManager,
    # oop_misc
    apply,
    MetaClsRepr,
)
from .os_tools import (
    pc_path,
    find_files_re,
    InDir,
    with_dir,
    maintain_cwd, maintain_directory, maintain_dir,
    get_paths_with_common_start,
    next_unique_name,
    nbytes_path,
)
from .properties import (
    alias, alias_to_result_of, alias_child, alias_key_of, alias_in,
    weakref_property_simple,
    simple_property, simple_tuple_property,
    elementwise_property,
    dict_with_defaults_property,
)
from .pytools import (
    printsource, displaysource,
    is_iterable,
    inputs_as_dict, _inputs_as_dict__maker,
    value_from_aliases,
    help_str, print_help_str, _help_str_paramdocs,
    indent_doclines, indent_paramdocs,
    pad_missing_format_keys, format_except_missing,
    replace_missing_format_keys, format_replace_missing,
)
from .sentinels import (
    Sentinel,
    UNSET, NO_VALUE, ATTR_UNSET, RESULT_MISSING,
)
from .supercomputer import (
    find_jobfiles, find_slurmfiles,
    SlurmOptionsDict, read_slurm_options,
    slurm_options_here, slurm_option_here,
)
from .timing import (
    Profile,
    PROFILE, profiling, print_profile,
    Stopwatch, TickingWatch,
    ProgressUpdater,
    TimeLimit,
)
from .trees import Tree
from .xarray_tools import (
    ## accessors ##
    XarrayAccessor, pcAccessor,
    # pcArrayAccessor, pcDatasetAccessor,  # <-- keep out of top-level namespace;
    # users can do pcAccessor.register(totype='array') or totype='dataset' instead.
    ## agg_stats ##
    xarray_aggregate,
    xarray_sum,
    xarray_stats,
    xarray_min, xarray_mean, xarray_median, xarray_max,
    xarray_std, xarray_rms,
    xarray_minimum, xarray_minimum_of_datavars,
    xarray_maximum, xarray_maximum_of_datavars,
    ## coords ##
    xarray_nondim_coords, xarray_dims_coords,
    xarray_assign_self_as_coord,
    xarray_fill_coords, xarray_index_coords,
    xarray_promote_index_coords, xarray_demote_index_coords,
    xarray_scale_coords, xarray_shift_coords, xarray_log_coords,
    xarray_is_sorted,
    # coord math
    xarray_get_dx_along, xarray_differentiate,
    ## dimensions ##
    is_iterable_dim,
    take_along_dimension, take_along_dimensions, join_along_dimension,
    xarray_rename, xarray_assign, xarray_promote_dim, xarray_ensure_dims,
    xarray_squeeze, xarray_squeeze_close, xarray_closeness,
    xarray_drop_unused_dims, xarray_drop_vars, xarray_popped,
    # broadcasting
    xarray_broadcastable_array, xarray_broadcastable_from_dataset,
    xarray_from_broadcastable, 
    # coarsen / windowing
    xarray_coarsened,
    ## grids ##
    xr1d, xrrange, xarray_range,
    XarrayGrid, xarray_grid,
    xarray_angle_grid,
    ## indexing ##
    xarray_isel, xarray_search, xarray_sel,
    xarray_where,
    xarray_map,
    xarray_argsort, xarray_sort_array_along, xarray_sort_dataset_along,
    xarray_cmin, xarray_cmax, xarray_varmin, xarray_varmax,
    xarray_at_min_of, xarray_at_max_of,
    xarray_min_coord_where, xarray_max_coord_where,
    ## io ##
    XarrayIoSerializable,
    xarray_save, xarray_load, xarray_mergeload,
    _xarray_save_prep,
    _xarray_coord_serializations, _xarray_coord_deserializations,
    ## masks ##
    xarray_mask,
    xarray_has_mask, xarray_store_mask, xarray_stored_mask, xarray_popped_mask,
    xarray_unmask,
    xarray_demask_from_mask, xarray_demask_from_ds,
    xarray_unmask_var, xarray_unmask_vars,
    ## misc ##
    xarray_copy_kw, xarray_with_data,
    xarray_as_array,
    xarray_vars_lookup, xarray_vars_lookup_with_defaults,
    xarray_where_finite,
    xarray_astypes, xarray_convert_types,
    xarray_object_coords_to_str,
    ## sci ##
    xarray_interp_inverse,
    xarray_gaussian_filter,
    xarray_polyfit, xarray_assign_polyfit_stddev, xarray_coarsened_polyfit,
    xarray_curve_fit, xarray_curve_eval, XarrayCurveFitter,
    XarrayLineFitter, xarray_line_fit,
    ## werr_stats ##
    xarray_werrmean,
    xarray_werr2pmstd, xarray_pmstd2werr, xarray_mean_pm_std,
    xarray_werradd, xarray_werrsub, xarray_werrmul, xarray_werrdiv,
)



"""
File Purpose: CopapicInputDeck - values from copapic.json file.
"""
import os

import numpy as np

from .copapic_dimensions import (
    CopapicDist, CopapicDistList,
    CopapicNeutral, CopapicNeutralList,
)
from .copapic_io_tools import read_copapic_json_file
from ...units import UnitsManagerPIC
from ...tools import (
    alias,
    DictlikeFromKeysAndGetitem,
)
from ...defaults import DEFAULTS


''' --------------------- CopapicInputDeck --------------------- '''

class CopapicInputDeck(DictlikeFromKeysAndGetitem):
    '''info from copapic.json file, in a class.
    Use CopapicInputDeck.from_file() to create an instance from a file.
    Units are "SI". 

    params: dict
        values of params
    distributions: dict or CopapicDistList
        dict of {distribution number: CopapicDist}, or an already-created CopapicDistList.
        if dict, will be internally stored as CopapicDistList.from_dict(distributions), instead.
    filename: str
        path to copapic.json file
        str will be internally stored as os.path.abspath(filename), instead.

    Examples:
        cpid = CopapicInputDeck.from_file('copapic.json')
        cpid[1]          # CopapicDist for distribution 1
        cpid['nz']       # number of grid points in z direction
        cpid.params      # dict of all global params
        cpid.dists       # CopapicDistList of all the distributions
        cpid.dists.get('e')    # get CopapicDist with name 'e'
        cpid.dists.get(0)      # get CopapicDist number 0 from this list
        cpid[2]['m']           # mass parameter for distribution 2
        cpid[0]['coll_rate']   # collision rate parameter for distribution 0
        cpid.keys()      # available keys
        cpid.items()     # available (key, value) pairs
        cpid[0].keys()   # available keys for distribution 0
        cpid.get('k', 7) # eid['k'] if it exists, else 7.
    '''
    dist_type = CopapicDist  # type for distributions
    dlist_type = CopapicDistList  # type for list of distributions
    neutral_type = CopapicNeutral  # type for neutral fluid
    nlist_type = CopapicNeutralList  # type for list of neutral fluids

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, params, distributions, *, filename=None, **kw_super):
        self.params = params
        self.distributions = distributions if isinstance(distributions, self.dlist_type) \
                                else self.dlist_type.from_dict(distributions)
        self._init_neutral_fluid(self.distributions)
        self.filename = None if filename is None else os.path.abspath(filename)
        super().__init__(**kw_super)

    dists = alias('distributions')

    def _init_neutral_fluid(self, dists):
        '''sets self.neutral and self.neutrals based on each distribution.
        dists: CopapicDistList
            list of distributions.
        '''
        ns = []
        if not isinstance(dists, self.dlist_type):
            raise TypeError(f'dists must be {self.dlist_type}, not {type(dists)}')
        if len(dists) == 0:
            raise ValueError('dists must have at least one distribution')
        for i, dist in enumerate(dists):
            if not isinstance(dist, self.dist_type):
                raise TypeError(f'dist must be {self.dist_type}, not {type(dist)}')
            pms = dist.params
            n = self.neutral_type(m=pms['neutralmass'], vth=pms['vneutralth'], 
                name=f'neutral_{i}', v0=pms['vneutral0'])
            ns.append(n)
        self.neutrals = self.nlist_type(ns)
        # self.neutral = self.neutral_type(m=-1, vth=[-1,-1,-1])
        # self.neutrals = self.nlist_type([self.neutral])

    @classmethod
    def from_file(cls, filename, *, dist_names=None, **kw__init):
        '''return CopapicInputDeck from copapic.json file.
        filename: str
            path to copapic.json file.
            abspath(filename) will be saved to result, for easier bookkeeping.
        dist_names: None or dict
            {N: name for distribution N} for any number of distributions in filename.
            None --> infer from file.
        '''
        deck = read_copapic_json_file(filename)
        if dist_names is None:
            dist_names = [n["name"] for n in deck["distributions"]]
            dist_names = {N: name for N, name in enumerate(dist_names)}
            dists = {N: n for N, n in enumerate(deck["distributions"])}
            for N, name in dist_names.items():
                if name is not None:
                    dists[N][f'name{N}'] = name 
        dist_dict = cls.dist_type.dists_from_dict(dists)
        params = {key: value for key, value in deck.items() if key != 'distributions'}
        return cls(params, dist_dict, filename=filename, **kw__init)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'with {len(self.dists)} distributions']
        if self.filename is not None:
            contents.append(f'filename={self.filename!r}')
        contents_str = ', '.join(contents)
        return f'{type(self).__name__}({contents_str})'

    # # # GETTING VALUES & ITERATING # # #
    def __getitem__(self, key):
        '''return self.params[key] if key in self.params, else self.dists[key].'''
        try:
            return self.params[key]
        except KeyError:
            pass  # handled below, to make clearer error message if failing.
        try:
            return self.dists[key]
        except (KeyError, IndexError, ValueError):
            errmsg = f'key {key!r} not in {type(self).__name__}.params or .dists'
            raise KeyError(errmsg) from None

    def keys(self):
        '''return tuple of keys to access self.dists or self.params'''
        return tuple(key for key in (*range(len(self.dists)), *self.params.keys()))

    # # # FILE PATH # # #
    @property
    def dirname(self):
        '''os.path.dirname(self.filename).
        if self.filename is None, raise FileNotFoundError.
        '''
        filename = self.filename
        if filename is None:
            raise FileNotFoundError('self.filename is None; cannot determine dirname.')
        return os.path.dirname(filename)
    
    @property
    def output_dir(self):
        '''return directory where output files are stored.
        '''
        title = self['title']
        if title is None:
            raise ValueError('title is None; cannot determine output_dir.')
        title = title.replace(' ', '')
        return os.path.join(self.dirname, title)

    # # # DIMENSIONS # # #
    def maindims(self):
        '''returns tuple of main dimensions in this input deck.
        inferred from self['nx'], 'ny', 'nz', and 'ndim_space'
        '''
        ndim = self['ndim']
        if ndim == 3:
            return ('x', 'y', 'z')
        elif ndim == 2:
            return ('x', 'y')  # by default, return ('x', 'y').
        else:
            raise NotImplementedError(f'{type(self).__name__}.maindims() when ndim_space={ndim}')

    def _get_nspace_simple(self, xstr):
        '''get "simple" number of cells in {x} dimension. == n{x} // nout_avg
        xstr should be 'x', 'y', or 'z'.
        See also: self._get_nspace.
        '''
        return self.get(f'n{xstr}', 1)

    def get_nspace(self, xstr):
        '''get number of cells in {x} dimension, accounting for nout_avg and nsubdomains.
        See also: Nx, Ny, Nz.
        '''
        result = self._get_nspace_simple(xstr)
        return result

    Nx = property(lambda self: self.get_nspace('x'),
                  doc='''actual total number of grid cells in the x dimension.''')
    Ny = property(lambda self: self.get_nspace('y'),
                  doc='''actual total number of grid cells in the y dimension.''')
    Nz = property(lambda self: self.get_nspace('z'),
                  doc='''actual total number of grid cells in the z dimension.''')

    def get_dspace(self, xstr):
        '''get d{x} [raw units] between output cells.
        '''
        return self.get(f'd{xstr}', 1)

    Dx = property(lambda self: self.get_dspace('x'),
                    doc='''dx between adjacent output cells [raw units]''')
    Dy = property(lambda self: self.get_dspace('y'),
                    doc='''dy between adjacent output cells [raw units]''')
    Dz = property(lambda self: self.get_dspace('z'),
                    doc='''dz between adjacent output cells [raw units]''')

    def get_coord(self, xstr):
        '''get coordinate array [raw units] for this dimension.
        xstr: str
            dimension name. 'x', 'y', or 'z'.
        '''
        dx = self.get_dspace(xstr)
        Nx = self.get_nspace(xstr)
        return np.arange(Nx) * dx

    def get_space_coords(self):
        '''return dict of {x: xcoord array} for x in self.maindims(), a subset of ('x', 'y', 'z').'''
        return {x: self.get_coord(x) for x in self.maindims()}

    def get_dspace_min(self):
        '''return min of self.Dx, self.Dy, self.Dz.'''
        return min(self.Dx, self.Dy, self.Dz)

    def get_dspace_max(self):
        '''return max of self.Dx, self.Dy, self.Dz.'''
        return max(self.Dx, self.Dy, self.Dz)

    # # # UNITS # # #
    def get_units_manager(self, *, u_l=None, u_t=None, u_n=None, ne_si=None, **kw_u_from_pic):
        '''return UnitsManagerPIC for this input deck
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return UnitsManagerPIC.from_pic(u_l=1.0)

    # # # QUASINEUTRAL? # # #
    def is_quasineutral(self):
        '''return whether this input deck is in quasineutral mode.
        as a proxy, guessing quasineutral iff any dist.is_hybrid() for dist in self.dists.
        '''
        return any(dist.is_hybrid() for dist in self.dists)


"""
File Purpose: tools related to multi-run analysis
"""
import os

from ..errors import FileContentsError
from ..tools import find_files_re, read_python_params_file


def canon_runs(dir=os.curdir, *, exclude=[], singles=False):
    '''returns dict of "canonical" runs within directory (and subdirectories).
    "canonical" if appearing within 'run' or 'runs' parameter of a _canon.txt file.

    keys are abspaths to directories containing the _canon.txt.
    values are lists of "canonical" runs within that directory.

    run paths within _canon.txt are interpreted from within the _canon.txt's directory.

    dir: str
        directory to search in.
    exclude: str, or list of strs
        exclude any subdirectories whose name equals one of these strings or re.fullmatch one of these strings.
        E.g. exclude='*[.]io' will exclude all subdirectories whose name ends with '.io';
            exclude='parallel' will exclude all subdirectories whose name equals 'parallel'.
    singles: bool
        if True, assert all run lists in result have length 1,
            and replace each runlist with runlist[0].
    '''
    canons = find_files_re('_canon[.]txt', dir=dir, exclude=exclude)
    result = {}
    for canon in canons:
        run_dir = os.path.dirname(canon)
        cparams = read_python_params_file(canon)
        runs = []
        if 'run' in cparams:
            runs.append(cparams['run'])
        if 'runs' in cparams:
            runs.extend(cparams['runs'])
        if len(runs) > 0:
            result.setdefault(run_dir, []).extend(runs)
    if singles:
        too_long = {k: v for k, v in result.items() if len(v) > 1}
        if len(too_long) > 0:
            errmsg = f"singles=True but _canon.txt specifies >1 run in some directories: {too_long}"
            raise FileContentsError(errmsg)
        for k, v in result.items():
            result[k] = v[0]
    # style: sort alphabetically by key
    result = {k: result[k] for k in sorted(result.keys())}
    return result



"""
File Purpose: various extensions to the Subplots class
"""
import numpy as np
import xarray as xr

from .images import XarrayImage
from .labels import XarraySubplotTitlesInferer, xarray_suptitle_plot_node, XarrayText
from .movies import MovieOrganizerNode, EmptyMovieNode
from .plot_settings import PlotSettings
from .plot_tools import calculate_vlims, plot_note
from .subplots import Subplots
from ..errors import (
    InputError, InputMissingError, InputConflictError,
    PlottingError, PlottingAmbiguityError,
)
from ..tools import (
    alias, alias_child,
    UNSET,
    wraplist, ndenumerate_nonNone, is_iterable,
    pcAccessor, take_along_dimensions, xarray_fill_coords,
    Tree,
)


@PlotSettings.format_docstring()
class MovieSubplots(MovieOrganizerNode, Subplots):
    '''MoviePlotNode for organizing a grid of subplots.
    Expects input 2D array (or nested list) of MoviePlotNodes,
        which were all created with init_plot=False
        (to allow MovieSubplots to create & assign the axes).
        For images, probably also want nodes to use add_colorbar=False
        (to allow MovieSubplots to create & assign the colorbar(s)).

    nodes: 2D array-like of MoviePlotNode or None objects
        the MoviePlotNodes associated with each subplot.
        None --> no MoviePlotNode for that subplot.
    name: str
        name for the MovieSubplots node; to be displayed in __repr__.
    parent: None or MoviePlotNode
        if provided, the parent of this node. None -> this node has no parent.
    init_plot: {init_plot}
    add_colorbars: {add_colorbars}

    kwargs which go to Subplots:
    axsize: {axsize}
    figsize: {figsize}
    {kw_subplots_adjust}
    {kw_share_axlike}
    {kw_share_ax}
    max_nrows_ncols: {max_nrows_ncols}
    squeeze: {squeeze}

    additional kwargs go to plt.subplots.

    Roughly equivalent functionality:
        subs = subplots(..., squeeze=False)  # creates a Subplots object
        axs = subs.axs  # the grid of axes in the Subplots
        msubs = MovieOrganizerNode()
        msubs_row0 = MovieOrganizerNode(parent=msubs)
        msubs_row1 = MovieOrganizerNode(parent=msubs)
        # ... as many rows as needed.
        MoviePlotNode(..., ax=axs[0,0], parent=msubs_row0)
        MoviePlotNode(..., ax=axs[0,1], parent=msubs_row0)
        # ...
        MoviePlotNode(..., ax=axs[1,0], parent=msubs_row1)
        MoviePlotNode(..., ax=axs[1,1], parent=msubs_row1)
        # ... as many nodes as needed.
        # --> while msubs wouldn't be a MovieSubplots object,
        #     it would still provide most of the same functionality,
        #     e.g. msubs.save(...) would properly animate & save movie!
    '''
    def __init__(self, nodes, name='', *, parent=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 add_colorbars=PlotSettings.default('add_colorbars'),
                 **kw_super):
        nodes = np.asanyarray(nodes)
        self.nodes = nodes
        if nodes.ndim != 2:
            raise InputError(f'nodes must be 2D, got shape {nodes.shape}')
        nrows, ncols = nodes.shape

        kw_super.update(nrows=nrows, ncols=ncols,
                init_plot=init_plot,
                add_colorbars=add_colorbars,
                name=name, parent=parent)
        # MovieOrganizerNode won't accidentally init_plot before axs are created,
        #   because it has obj=None so it never calls init_plot during __init__.
        super().__init__(**kw_super)

        if self.plot_settings['init_plot']:
            self.init_plot()

    def init_plot(self):
        '''plot for the first time: call init_plot on all nodes,
        and organize into a nice tree structure which can be indexed like an array.

        This is fundamentally different from MoviePlotNode.init_plot's usual purpose,
            since this is about calling init_plot on the nodes, not on self.obj.
            [TODO] should this function be renamed, for clarity?
        '''
        # checks: skip the usual checks; init_plot on nodes, not self.obj.
        # but, do check if self.init_plot was already called; don't want to call it twice!
        if getattr(self, '_called_init_plot', False):
            errmsg = (f'init_plot was already called; cannot call it again. '
                      f'For {type(self).__name__}(name={self.name!r}) with id={hex(id(self))}')
            raise PlottingAmbiguityError(errmsg)
        # bookkeeping
        name_append = f' {self.name}' if self.name else ''
        axs = self.axs
        # call init_plots on nodes, set up row organizers, & create appropriate node tree.
        for i, row in enumerate(self.nodes):
            row_organizer = MovieOrganizerNode(name=f'(row={i}){name_append}', parent=self)
            for j, node in enumerate(row):
                if node is None:
                    node = EmptyMovieNode()
                else:
                    node._ax_init = axs[i,j]
                    node.init_plot()
                node.parent = row_organizer
        # formatting
        self.remove_redundant_labels()
        self.hide_empty_axes()
        add_colorbars = self.plot_settings['add_colorbars']
        if add_colorbars:
            if self.plot_settings.get('add_colorbar', False):
                raise InputConflictError('Can add_colorbars or add_colorbar, but not both!')
            self.colorbars(mode=add_colorbars)

    # # # INHERITANCE FUNNY BUSINESS # # #
    # we want this object to behave sometimes like MoviePlotNode, sometimes like Subplots.
    # Those two classes have mostly non-overlapping methods, but there is some overlap.

    # In particular, both define methods for indexing. Here we want to use Tree-like indexing!
    __getitem__ = Tree.__getitem__
    __iter__ = Tree.__iter__
    # [TODO] if there's a "cleaner" way to do this, that might be nice...

    # # # DISPLAY # # #
    DEFAULT_TREE_SHOW_DEPTH = 2   # use self.display(show_depth=...) to see with a different depth.


@pcAccessor.register('subplots')
@PlotSettings.format_docstring()
class XarraySubplots(MovieSubplots):
    '''grid of subplots from xarray data. Can be animated as a movie!

    xarray: xarray.DataArray or xarray.Dataset
        the array to plot. up to 5D (row, col, x, y, t).
        (Note that row & col dims should be "not too long" otherwise plot will be very large)
        if dataset, up to 4D; will be converted to DataArray with new dim named 'variable'.
        internally, will store xarray_fill_coords(xarray) to utilize coordless dims' indices in titles.
    row: None or str
        dimension to plot ACROSS rows.
        None --> subplots will have ncols=1 (nothing varies across a row --> only 1 column).
        E.g. if row=='fluid', then the i'th COLUMN will be fluid i.
        [TODO] infer row directly from xarray, somehow?
    col: None or str
        dimension to plot DOWN columns.
        None --> subplots will have nrows=1 (nothing varies down a column --> only 1 row).
        E.g. if col=='component', then the j'th ROW will be component j.
    wrap: None or int
        wrap row or col dimension after this many images.
        can only be provided if provided row or col but not both.
        E.g. if row=='fluid', wrap=4, and input has 10 fluids, will make rows of 4, 4, 2 images.
            Any remaining "empty" plots will have their axes hidden.
    x, y: None or str
        dimension name for the x and y axes of an individual subplot.
        None --> infer from the xarray coords. See also: DEFAULTS.PLOT.DIMS_INFER
    t: None or str
        dimension name for the time axis (for movies).
        None --> infer from the xarray coords. See also: DEFAULTS.PLOT.DIMS_INFER

    vmin, vmax: None, scalar value, or array-like
        if provided, tells value for vmin, vmax for all subplots, ignoring share_vlims and robust.
        if providing only one, still use share_vlims and robust for the other
            (e.g. if provided vmax but vmin=None, use share_vlims and robust when deciding on vmin).
        if array-like vmin, use vmin=vmin[i][j] for plot in i'th row and j'th column. Similar for vmax.
            (doesn't squeeze. e.g., if only 1 row exists, because row=None, then use vmin[0][j])
            (if non-None wrap, vlims shape should correspond to the shape after wrapping.)
            (use None limit to instead use share_vlims and robust for that subplot.
                E.g. vmin=[[0, 2, None]], share_vlims='row', robust=10, for subplots with 1 row, 3 cols,
                will use vmin=0 for first plot, 2 for second plot,
                    and vmin=10th percenticle across all values in all three plots, for the third plot.)
    share_vlims: {share_vlims}
    robust: {robust}

    axsize: {axsize}
    aspect: {aspect}
    layout: {layout}

    suptitle: {suptitle}
    suptitle_y: {suptitle_y}
    suptitle_font: {suptitle_font}
    suptitle_kw: {suptitle_kw}
    suptitle_width: {suptitle_width}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    subplot_title_width: {subplot_title_width}

    rtitle: {rtitle}
    rtitle_kw: {rtitle_kw}
    ttitle: {ttitle}
    ttitle_kw: {ttitle_kw}

    cax_mode: {cax_mode}
    colorbar_kw: {colorbar_kw}
    add_colorbars: {add_colorbars}

    additional kwargs go to other places [TODO][DOC] fill in the details.
    '''
    def __init__(self, xarray, row=None, col=None, *, wrap=None, x=None, y=None, t=None,
                 fig=None,
                 vmin=None, vmax=None,
                 robust=PlotSettings.default('robust'),  # <- could go in **kw, but explicit is nice.
                 share_vlims=PlotSettings.default('share_vlims'),
                 title=PlotSettings.default('title'),
                 title_y=PlotSettings.default('title_y'),
                 suptitle=PlotSettings.default('suptitle'),
                 suptitle_y=PlotSettings.default('suptitle_y'),
                 rtitle=PlotSettings.default('rtitle'),
                 ttitle=PlotSettings.default('ttitle'),
                 aspect=PlotSettings.default('aspect'),
                 layout=PlotSettings.default('layout'),
                 cax_mode=PlotSettings.default('cax_mode'),
                 axsize=PlotSettings.default('axsize'),
                 add_colorbars=PlotSettings.default('add_colorbars'),
                 **kw):
        name = '' if (getattr(xarray, 'name', None) is None) else xarray.name
        if isinstance(xarray, xr.Dataset):
            xarray = xarray.to_array(dim='variable')
        xarray = xarray_fill_coords(xarray, need=[row, col, x, y, t])
        self.array = xarray
        # grid of xarray DataArrays, based on self.array, row, col, and wrap.
        arrs = self._get_xarray_arr(self.array, row=row, col=col, wrap=wrap)
        self.arrs = arrs
        # settings -- update kw with settings which are explicitly in function call signature.
        kw.update(title=title, title_y=title_y,
                  suptitle=suptitle, suptitle_y=suptitle_y,
                  rtitle=rtitle, ttitle=ttitle,
                  aspect=aspect, layout=layout, cax_mode=cax_mode,
                  add_colorbars=add_colorbars, axsize=axsize,
                  share_vlims=share_vlims, robust=robust)
        init_plot_settings = PlotSettings(**kw)
        # [TODO] infer row & col plot dimensions if not provided.
        self.row_plot_dim = row
        self.col_plot_dim = col
        self._wrap_input = wrap
        # vmin and vmax are handled separately in case we are sharing vlims or using robust:
        vlims_arr = self._get_vlims_arr(arrs, vmin=vmin, vmax=vmax, share_vlims=share_vlims, robust=robust)
        # create array of XarrayImage nodes (but don't init_plot or add_colorbar yet!)
        images = np.full(arrs.shape, None, dtype=object)
        for idx, arr in ndenumerate_nonNone(arrs):
            vmin, vmax = vlims_arr[idx]
            image = XarrayImage(arr, t=t, x=x, y=y,
                                vmin=vmin, vmax=vmax,
                                **{**kw,  # override input kw for these options here:
                                'init_plot': False, 'add_colorbar': False,
                                })
            # ensure arr is 2D or 3D (with an inferred t_plot_dim); else crash with helpful message.
            if not image._has_valid_plot_dims():
                errmsg = (f'failed to infer plot dimensions; image array has dims {image.array.dims!r}.\n'
                          'Most likely, you forgot to specify row, col, x, y, and/or t.')
                raise PlottingAmbiguityError(errmsg)
            images[idx] = image
        # infer subplot titles where not provided
        self.images = images  # <-- must do this before accessing self.t_plot_dim.
        kw_infer_titles = init_plot_settings.get_mpl_kwargs('pc.XarraySubplotTitlesInferer')
        self.titles_inferer = XarraySubplotTitlesInferer(self.array, t=self.t_plot_dim,
                                        row=self.row_plot_dim, col=self.col_plot_dim,
                                        **kw_infer_titles)  # titles_inferer might be used later even if not now.
        if (ttitle is not UNSET) and title is not PlotSettings.default('title'):
            raise InputConflictError('Cannot provide both ttitle and title at the same time!')
        if init_plot_settings['add_labels']:
            if ttitle is UNSET:
                if title is PlotSettings.default('title'):
                    for idx, image in ndenumerate_nonNone(images):
                        image.plot_settings['title'] = self.titles_inferer.infer_title()
            else:  # provided ttitle
                font = {'fontfamily': init_plot_settings['title_font'],
                        **init_plot_settings.get('title_kw', default={})}['fontfamily']
                ttitle_kw = {'fontfamily': font, **init_plot_settings['ttitle_kw']}
                for image in images[0, :]:
                    image.plot_settings['title'] = ttitle
                    image.plot_settings['title_kw'] = ttitle_kw
                for image in images[1:, :].flat:
                    image.plot_settings['title'] = None
        # init super
        super().__init__(images, fig=fig, name=name, **kw)  # [TODO] does non-None fig work here?
        # <- rtitle handled by self.init_plot()

    # # # INIT PLOT # # #
    def init_plot(self):
        '''plot for the first time: call init_plot on all nodes,
        and organize into a nice tree structure which can be indexed like an array.

        This is fundamentally different from MoviePlotNode.init_plot's usual purpose,
            since this is about calling init_plot on the nodes, not on self.obj.
            [TODO] should this function be renamed, for clarity?
        '''
        super().init_plot()
        # add suptitle after super().init_plot() adds children nodes for subplots.
        if self.plot_settings['add_labels']:
            self.plot_suptitle()
            self.plot_rtitles()

    # # # PROPERTIES # # #
    images = alias('nodes')
    image0 = property(lambda self: self.images[0,0], doc='''top-left image; images[0,0].''')
    x_plot_dim = alias_child('image0', 'x_plot_dim')
    y_plot_dim = alias_child('image0', 'y_plot_dim')
    t_plot_dim = alias_child('image0', 't_plot_dim')

    @property
    def isels(self):
        '''np.array (dtype=object) of (None or dict of index details), one for each subplot.
        E.g. full_array.isel(self.isels[1,4]) == self.images[1,4].array.
        '''
        rowdim = self.row_plot_dim
        coldim = self.col_plot_dim
        if (rowdim is None) and (coldim is None):
            raise NotImplementedError('[TODO] isels when row and col are both None')
        if (rowdim is not None) and (coldim is None):
            Lrowdim = self.array.sizes[rowdim]
            idxarr = np.empty(Lrowdim, dtype=object)
            for i in range(Lrowdim):
                 idxarr[i] = {rowdim: i}
            idxarr = xr.DataArray(idxarr, dims=[rowdim])
        elif (rowdim is None) and (coldim is not None):
            Lcoldim = self.array.sizes[coldim]
            idxarr = np.empty(Lcoldim, dtype=object)
            for i in range(Lcoldim):
                 idxarr[i] = {coldim: i}
            idxarr = xr.DataArray(idxarr, dims=[coldim])
        elif (rowdim is not None) and (coldim is not None):
            Lrowdim = self.array.sizes[rowdim]
            Lcoldim = self.array.sizes[coldim]
            idxarr = np.empty((Lrowdim, Lcoldim), dtype=object)
            for i in range(Lrowdim):
                for j in range(Lcoldim):
                    idxarr[i, j] = {rowdim: i, coldim: j}
            idxarr = xr.DataArray(idxarr, dims=[rowdim, coldim])
        result = self._get_xarray_arr(idxarr, row=rowdim, col=coldim, wrap=self._wrap_input)
        as_numpy = np.empty(result.shape, dtype=object)
        for idx, arr in np.ndenumerate(result):
            as_numpy[idx] = None if arr is None else arr.item()
        return as_numpy

    def rightmost_images(self, *, as_idx=False, missing_ok: True):
        '''list of rightmost existing image (i.e., non-None) in each row.
        as_idx: bool
            whether to return indices of images within self.images, instead of image objects.
        missing_ok: bool
            whether it is okay for some row to have no images (result will be None in those rows).
        '''
        result = []
        for irow, row in enumerate(self.images):
            for minus_icol, image in enumerate(row[::-1]):
                icol = len(row) - minus_icol - 1
                if image is not None:
                    result.append((irow, icol) if as_idx else image)
                    break
            else:
                if missing_ok:
                    result.append(None)
                else:
                    raise PlottingError(f'row {irow} has no images, and missing_ok=False!')
        return result

    # # # SUPTITLE # # #
    def plot_suptitle(self):
        '''adds suptitle (as a MovieText node)
        raise PlottingAmbiguityError if suptitle already plotted
            (this prevents having multiple suptitle nodes).
        '''
        if hasattr(self, 'suptitle_node'):
            errmsg = (f'suptitle already plotted; cannot plot it again. '
                      f'For {type(self).__name__}(name={self.name!r}) with id={hex(id(self))}')
            raise PlottingAmbiguityError(errmsg)
        suptitle = self.plot_settings['suptitle']
        if suptitle is UNSET:
            suptitle = self.titles_inferer.infer_suptitle()
        if isinstance(suptitle, str):
            kw_suptitle = {**self.plot_settings.kw, 'init_plot': True}
            suptitle_node = xarray_suptitle_plot_node(self.array, suptitle, t=self.t_plot_dim,
                                                      fig=self.fig, parent=self, **kw_suptitle)
            self.suptitle_node = suptitle_node  # <-- for bookkeeping/debugging

    # # # RTITLES # # #
    def plot_rtitles(self):
        '''adds rtitles (as MovieText node objects) via plot_note() on right-hand-side of right-most images.
        (only if self.plot_settings['rtitle'] is not UNSET.)
        '''
        rtitle = self.plot_settings['rtitle']
        if rtitle is UNSET:
            return
        # else, rtitle was provided.
        rimages = self.rightmost_images(as_idx=False, missing_ok=True)
        font = {'fontfamily': self.plot_settings['title_font'],
                **self.plot_settings.get('title_kw', default={})}['fontfamily']
        rtitle_kw = {'fontfamily': font, **self.plot_settings['rtitle_kw']}
        for image in rimages:
            if image is None:
                continue
            rtitle_text = plot_note(rtitle, ax=image.ax, **rtitle_kw)
            rtitle_node = XarrayText(image.array, rtitle_text, parent=image)
            image.rtitle_node = rtitle_node  # <-- for bookkeeping/debugging

    # # # ARRS FROM INPUTS # # #
    def _get_xarray_arr(self, xarray, row=None, col=None, wrap=None):
        '''return array of xarray.DataArray objects, organized into a grid.
        E.g. when wrap=None, result[i,j] = self.array.isel({row: i, col: j}).
        see help(type(self)) for details on row, col, wrap.
        '''
        if wrap is None:
            # [col, row] order agrees with definition of row & col kwargs.
            result = take_along_dimensions(xarray, [col, row])
        else:  # wrap is not None
            if row is None and col is None:
                raise InputMissingError(f'row or col, when wrap is not None. (got wrap={wrap!r})')
            if row is not None and col is not None:
                errmsg = ('wrap can only be provided if provided row or col but not both. '
                          f'Got wrap={wrap!r}, row={row!r}, col={col!r}')
                raise InputConflictError(errmsg)
            if row is not None:
                result = take_along_dimensions(xarray, row)
                result = wraplist(result, wraprow=wrap)
            else: # col is not None:
                result = take_along_dimensions(xarray, col)
                result = wraplist(result, wrapcol=wrap)
        return result

    # # # VLIMS # # #
    _calculate_vlims = staticmethod(calculate_vlims)

    @PlotSettings.format_docstring(ntab=2)
    def _get_vlims_arr(self, arrs=None, *, vmin=None, vmax=None, share_vlims=False, robust=UNSET):
        '''return array of vmin, vmax, for each array in arrs.
        arrs: None or array of arrays.
            if None, use self.arrs.
        vmin, vmax: None, value, or array-like with same shape as `arrs`
            if provided, use this value for vmin, vmax for all subplots, ignoring share_vlims and robust.
            See help(type(self)) for more details.
        share_vlims: {share_vlims}
        robust: {robust}
        '''
        # [EFF] handle simple edge case first (both non-None scalars):
        if not is_iterable(vmin) and not is_iterable(vmax) and vmin is not None and vmax is not None:
            result = np.empty(arrs.shape, dtype=object)
            for idx, _ in np.ndenumerate(result): result[idx] = (vmin, vmax)
            return result
        # bookkeeping
        if robust is UNSET:
            robust = PlotSettings.get_default('robust')
        if arrs is None:
            arrs = self.arrs 
        if not is_iterable(vmin):
            vmin = np.full(arrs.shape, vmin, dtype=object)
        elif not hasattr(vmin, 'ndim'):
            vmin = np.asarray(vmin)
        if not is_iterable(vmax):
            vmax = np.full(arrs.shape, vmax, dtype=object)
        elif not hasattr(vmax, 'ndim'):
            vmax = np.asarray(vmax)
        def has_None_where_other_doesnt(array0, array1):
            return (None in array0) and any(a0 is None and a1 is not None for a0, a1 in zip(array0.flat, array1.flat))
        result = np.full(arrs.shape, None, dtype=object)  # array of tuples.
        # compute results
        if share_vlims is True or share_vlims=='all':
            if True:  # I just wanted the indentation to match code below...
                avals = [a.values for a in arrs.flat if a is not None]
                kw_vlims = dict(vmin=None if has_None_where_other_doesnt(vmin, arrs) else 0,
                                vmax=None if has_None_where_other_doesnt(vmax, arrs) else 1)
                rmin, rmax = self._calculate_vlims(avals, robust=robust, **kw_vlims, expand_flat=False)
                for idx, _ in np.ndenumerate(result):
                    result[idx] = (rmin if vmin[idx] is None else vmin[idx],
                                   rmax if vmax[idx] is None else vmax[idx])
        elif share_vlims == 'row':
            for i, row in enumerate(arrs):
                avals = [a.values for a in row if a is not None]
                kw_vlims = dict(vmin=None if has_None_where_other_doesnt(vmin[i], row) else 0,
                                vmax=None if has_None_where_other_doesnt(vmax[i], row) else 1)
                rmin, rmax = self._calculate_vlims(avals, robust=robust, **kw_vlims, expand_flat=False)
                for j, _ in enumerate(row):
                    result[i,j] = (rmin if vmin[i,j] is None else vmin[i,j],
                                   rmax if vmax[i,j] is None else vmax[i,j])
        elif share_vlims == 'col':
            for j in range(arrs.shape[1]):
                col = arrs[:, j]
                avals = [a.values for a in col if a is not None]
                kw_vlims = dict(vmin=None if has_None_where_other_doesnt(vmin[:,j], col) else 0,
                                vmax=None if has_None_where_other_doesnt(vmax[:,j], col) else 1)
                rmin, rmax = self._calculate_vlims(avals, robust=robust, **kw_vlims, expand_flat=False)
                for i in range(arrs.shape[0]):
                    result[i,j] = (rmin if vmin[i,j] is None else vmin[i,j],
                                   rmax if vmax[i,j] is None else vmax[i,j])
        elif not share_vlims:
            for idx, arr in np.ndenumerate(arrs):
                if arr is None:
                    result[idx] = (None, None)
                    continue
                avals = arr.values
                kw_vlims = dict(vmin=None if (vmin[idx] is None and arr is not None) else 0,
                                vmax=None if (vmax[idx] is None and arr is not None) else 1)
                rmin, rmax = self._calculate_vlims(avals, robust=robust, **kw_vlims, expand_flat=False)
                result[idx] = (rmin if vmin[idx] is None else vmin[idx],
                               rmax if vmax[idx] is None else vmax[idx])
        else:
            raise InputError(f'invalid share_vlims: {share_vlims}, expected True, False, "all", "row", or "col".')
        return result


    # # # MISC CONVENIENT METHODS # # #
    def scatter_max(self, **kw_scatter):
        '''use XarrayScatter to plt.scatter() a marker at argmax of each subplot image's array.
        animatable (e.g. different max for each frame); for each image, image.add_child(scatter result).
        default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MAX}
        returns np.array with same shape as self, containing the results of XarrayScatter calls.
        '''
        results = np.full(self.shape, None, dtype=object)
        for idx, image in ndenumerate_nonNone(self.images):
            results[idx] = image.scatter_max(**kw_scatter)
        return results

    def scatter_min(self, **kw_scatter):
        '''use XarrayScatter to plt.scatter() a marker at argmin of each subplot image's array.
        animatable (e.g. different max for each frame); for each image, image.add_child(scatter result).
        default style: {**DEFAULTS.PLOT.SCATTER_STYLE, **DEFAULTS.PLOT.SCATTER_MAX}
        returns np.array with same shape as self, containing the results of XarrayScatter calls.
        '''
        results = np.full(self.shape, None, dtype=object)
        for idx, image in ndenumerate_nonNone(self.images):
            results[idx] = image.scatter_min(**kw_scatter)
        return results

    def scatter_where(self, condition, **kw_scatter):
        '''use XarrayScatter to plt.scatter() markers where condition is True.
        animatable (e.g. different max for each frame); for each image, image.add_child(scatter result).

        condition: callable or DataArray of bools
            callable --> use condition(arr) for each image in self, where arr=image.array.
            DataArray --> use condition.isel(ii) for each image in self,
                        where ii is the corresponding dict of indices from self.isels.

        returns np.array with same shape as self, containing the results of XarrayScatter calls.

        Example:
            xsubs = array.pc.image(...)
            xsubs.scatter_where(lambda arr: arr > 7.5)  # markers at all values larger than 7.5
            xsubs.save(filename)  # save animation to filename (if xsubs.t_plot_dim is not None)
        '''
        results = np.full(self.shape, None, dtype=object)
        for idx, isel in ndenumerate_nonNone(self.isels):
            image = self.images[idx]
            if image is None:
                continue
            if callable(condition):
                where = condition  # image.scatter_where handles the rest.
            else:
                where = condition.isel(isel, missing_dims='ignore')
            results[idx] = image.scatter_where(where, **kw_scatter)
        return results



"""
File Purpose: tools to create grids with xarray
"""
import numpy as np
import xarray as xr

from ..iterables import is_iterable
from ..properties import alias
from ..docs_tools import format_docstring
from ..sentinels import UNSET
from ...errors import (
    InputError, InputMissingError, InputConflictError,
    DimensionSizeError,
)


''' --------------------- xr1d --------------------- '''

def xr1d(arraylike, name, *, coords=dict(), **kw_data_array):
    '''create a 1D xarray.DataArray from arraylike, using name for array, dim, & coord.
    Equivalent: xr.DataArray(arraylike, coords={name: arraylike}, name=name, dims=name)

    name: str
        name of the array. Result name, dim, and coord will all use this name.
    coords: dict
        if provided, update default coords using this dict.
        E.g. xr1d([1,2,3], 'x', coords={'y': 7}) coords will be {'x': [1,2,3], 'y': 7}.
        E.g. xr1d([1,2,3], 'x', coords={'x': [1,2,3]+10}) coord for 'x' will be [11,12,13].

    Example: xr1d([1,7,3], 'x')
        <xarray.DataArray 'x' (x: 3)> Size: 24B
        array([1, 7, 3])
        Coordinates:
          * x        (x) int64 24B 1 7 3
    '''
    if not isinstance(name, str):
        raise InputError(f'expected str name, but got name={name!r}')
    coords = {name: arraylike, **coords}
    return xr.DataArray(arraylike, dims=[name], coords=coords, name=name, **kw_data_array)


''' --------------------- xarray_range --------------------- '''

def xrrange(range_info, name='range', *, coords=dict(), **kw_data_array):
    '''create a 1D xarray.DataArray of np.arange(*range_info), using name for array, dim, & coord.
    if range_info is a single number, use np.arange(range_info) instead.
    Equivalent: xr.DataArray(np.arange(*range_info), coords={name: np.arange(*range_info)}, ...)
    
    range_info: int or iterable of 1, 2, or 3 ints (or None)
        (will be passed directly to np.arange.)
        1 int provided --> stop
        2 ints provided --> start, stop
        3 ints provided --> start, stop, step
    name: str
        name of the range. Result name, dim, and coord will all use this name.
    coords: dict
        if provided, update default coords using this dict.
        E.g. xrrange(4, 'x', coords={'y': 7}) coords will be {'x': np.arange(4), 'y': 7}.
        E.g. xrrange(4, 'x', coords={'x': np.arange(4)+10}) coord for 'x' will be [10,11,12,13].

    Example: xrrange(4, 'x')  # ==xr.DataArray(np.arange(4), coords={'x': np.arange(4)}, name='x')
        <xarray.DataArray 'x' (x: 4)> Size: 32B
        0 1 2 3
        Coordinates:
          * x        (x) int64 32B 0 1 2 3
    '''
    if is_iterable(range_info):
        if len(range_info) == 0 or len(range_info) > 3:
            raise ValueError(f'expected range_info to be 1, 2, or 3 ints, but got len={len(range_info)}')
    else:
        range_info = [range_info]
    arange = np.arange(*range_info)
    return xr1d(arange, name, coords=coords, **kw_data_array)

xarray_range = xrrange  # alias to xrrange


''' --------------------- XarrayGrid --------------------- '''

class XarrayGrid():
    '''class to help with making xarray.DataArray grids.
    call self (e.g.: self()) or self.grid() to get the grid as an xarray.DataArray.
    
    intended to be "immutable" after creation;
        altering self.min, self.max or other attrs may produce unexpected results.

    Inputs and outputs will always be in linear space (even if logspace=True)!


    --- Params related to lims ---

        min: None, number, DataArray, or callable
            min value (or array of min values)
            None --> infer from max and other params.
            callable f --> infer min = f(max)
        max: None, number, DataArray, or callable
            max value (or array of max values)
            None --> infer from min and other params.
            callable f --> infer max = f(min)
        span: None or number
            max - min. None = "unspecified".
            only allowed if min or max is None.
        ratio: None or number
            max / min. None = "unspecified".
            only allowed if min or max is None.

    --- Params related to grid size & spacing ---
        N: None or number
            number of points in grid. None = "unspecified"
        step: None or number
            step size between points (in linear space). None = "unspecified"
            if provided, implies logspace=False.
        logspace: None or bool
            whether the grid will be evenly spaced in log-space.
            None --> True if provided logstep; False if provided N or step.
            True --> must have provided N or logstep.
            False --> must have provided N or step.
        logstep: None or number
            step size between points, in log-space (base 10). None = "unspecified"
            if provided, implies logspace=True.
        inclusive: tuple of 2 bools
            whether to include min and max in the output.
        reverse: bool
            whether to reverse the output when creating grid.
            Equivalent to using result.isel(grid_dim=slice(None, None, -1)),
                but replacing 'grid_dim' with the appropriate name.

    --- Other params ---
        name: str
            result.name; also result.assign_coords({name:result})
            (ops with result keep coords[name], e.g. result + 10, or result * other_array)
        dim: None or str
            grid dimension. For scalar min & max, dim=name works great & intuitively.
            However, when self.array_lims, dim cannot equal name,
                because result will be n+1 dimensional (n=ndims from max-min),
                so setting result.assign_coords({name:result}) would fail if name is a dim.
            None --> dim=name if not self.array_lims else '{name}_dim'
            str --> use dim.format(name=name).
        math_coord: bool or str
            whether to add coord related to the grid-producing math, if self.array_lims.
            default name '{name}_normed', with values a 1D array from 0 to 1, such that:
                if linspace, result = min + (max - min) * coord
                if logspace, result = 10**(log10(min)+(log10(max)-log10(min))*coord)
            str --> use this coord, after doing math_coord.format(name=name).
        N_min: UNSET, None or number
            if provided, tells minimum allowed number of points in grid.
            None --> no minimum. UNSET --> use type(self).N_min (default=2)
        N_max: UNSET, None or number
            if provided, tells maximum allowed number of points in grid.
            None --> no maximum. UNSET --> use type(self).N_max (default=1e8)

    --- notes about result coords ---
        result.name = `name`
        result.coords[name] == result.
        result.dims = [dim, *dims from min and/or max (broadcasted appropriately)]
        If inputs are scalars:
            dim = `name`, unless explicitly provided `dim`.
                Either way, then use dim.format(name=name).
            result.dims = [dim]
            if dim == name, no additional coords added to result,
            otherwise, add coord: result[name] = result.
        If inputs are arrays, 
            dim = '{name}_dim' unless explicitly provided. 
                Either way, then use dim.format(name=name).
            result.dims = [dim, *dims from min and/or max]
            dim == name is NOT allowed.
            always add coord: result[name] = result
            if math_coord=True, may also add coord '{name}_normed'

    --- bookkeeping - values computed during init ---
    (__init__ computes these. Here is some documentation about what they mean.)

        self.step_param: 'N', 'step', or 'logstep'
            tells which step-related param was provided.
            (exactly 1 of these must be provided, else will crash.)
        self.array_lims: bool
            tells whether min or max is an array.
    '''
    DEFAULT_DIM = '{name}_dim'  # used when inputs are arrays, unless dim explicitly provided.
    NORMED_COORD = '{name}_normed'

    N_min = 2
    N_max = 1e8

    def __init__(self, min=None, max=None, N=None, name='grid', *, step=None,
                 span=None, ratio=None,
                 logspace=None, logstep=None,
                 inclusive=(True, True), reverse=False,
                 dim=None, math_coord=False,
                 N_min=UNSET, N_max=UNSET):
        self.init_step_params(N=N, step=step,
                              logspace=logspace, logstep=logstep,
                              inclusive=inclusive)
        self.reverse = reverse
        if N_min is not UNSET: self.N_min = N_min
        if N_max is not UNSET: self.N_max = N_max
        self.init_lims(min=min, max=max, span=span, ratio=ratio)
        self.init_labels(name=name, dim=dim, math_coord=math_coord)

    def init_step_params(self, *, N, step, logspace, logstep, inclusive):
        '''initialize N, step, logspace, logstep, inclusive from kwargs.
        Also sets self.step_param = 'N', 'step', or 'logstep'
        '''
        kw_step_input = dict(N=N, step=step, logstep=logstep)
        provided = [k for k, v in kw_step_input.items() if v is not None]
        if len(provided) == 0:
            raise InputMissingError('must provide one of: N, step, logstep.')
        if len(provided) >= 2:
            errmsg = f'expected only one of: N, step, logstep; got values for: {provided}'
            raise InputConflictError(errmsg)
        step_param = provided[0]
        self.step_param = step_param
        # bookkeeping for logspace
        if logspace is None:  # then, infer it:
            if step_param in ['N', 'step']:
                logspace = False
            elif step_param in ['logstep']:
                logspace = True
            else:
                assert False, 'coding error if reached this line'
        elif (logspace) and (step_param == 'step'):
            raise InputConflictError('logspace=True incompatible with step. Use N or logstep instead.')
        elif (not logspace) and (step_param == 'logstep'):
            raise InputConflictError('logspace=False incompatible with logstep. Use N or step instead.')
        # setting params internally
        self.N = N
        self.step = step
        self.logspace = logspace
        self.logstep = logstep
        self.inclusive = inclusive

    def init_lims(self, *, min, max, span, ratio):
        '''initialize self.min, max, span, and ratio. Also calls callables.
        Also sets self.array_lims = bool (True if min or max is an array).
        Also, if provided min and max, but np.any(min > max), raise InputConflictError.
        '''
        if callable(min):
            min = min(max)
        if callable(max):
            max = max(min)
        if min is None and max is None: raise InputMissingError('must provide min or max.')
        if span is None and ratio is None:
            if min is None: raise InputMissingError('must provide min, span, or ratio.')
            if max is None: raise InputMissingError('must provide max, span, or ratio.')
        if min is not None and max is not None:
            check = min > max  # <-- bad if True anywhere...
            if np.any(check):
                if getattr(check, 'ndim', 0) > 0:
                    errmsg = f'min > max (at {int(check.sum())} points, out of {check.size}).'
                else:
                    errmsg = f'min > max... cannot make grid.'
                raise InputConflictError(errmsg)
        if min is not None and max is not None:
            if span is not None: 
                raise InputConflictError('cannot provide span when providing both min & max.')
            if ratio is not None:
                raise InputConflictError('cannot provide ratio when providing both min & max.')
        if span is not None and ratio is not None:
            raise InputConflictError('cannot provide both span and ratio.')
        if getattr(span, 'ndim', 0) > 0:
            raise InputError(f'non-scalar span; got span=\n{span}')
        if getattr(ratio, 'ndim', 0) > 0:
            raise InputError(f'non-scalar ratio; got ratio=\n{span}')
        self.min = min
        self.max = max
        self.span = span
        self.ratio = ratio
        self.array_lims = (isinstance(min, xr.DataArray) or isinstance(max, xr.DataArray))
        if self.array_lims:
            if self.step_param=='step' and self.span is None:
                errmsg = ("array_lims=True when providing 'step', only allowed if provided span too.\n"
                          "Either provide 'span' (and leave ratio=None, and min=None or max=None),\n"
                          "or provide 'N' (and leave step=None).")
                raise InputConflictError(errmsg)
            if self.step_param=='logstep' and self.ratio is None:
                errmsg = ("array_lims=True when providing 'logstep', only allowed if provided ratio too.\n"
                          "Either provide 'ratio' (and leave span=None, and min=None or max=None),\n"
                          "or provide 'N' (and leave logstep=None).")
                raise InputConflictError(errmsg)
        
    def init_labels(self, *, name, dim, math_coord):
        '''initialize self.name, dim, and math_coord, using defaults if needed.
        result depends on whether self.min and max are arrays (i.e. self.array_lims).
        See help(type(self)) for more details.
        '''
        self.name = name
        if dim is None:
            dim = self.DEFAULT_DIM if self.array_lims else name
        if self.array_lims and dim == name:
            errmsg = (f'name==dim (=={dim!r}) when self.array_lims==True, is not allowed.\n'
                      'grid result[name] == result.values which will vary across array_lims dims,\n'
                      'so name cannot equal a dimension name, since result[name] will be >1D.')
            raise InputConflictError(errmsg)
        dim = dim.format(name=name)
        self.dim = dim
        if math_coord == True:
            if self.varied_step:
                math_coord = self.NORMED_COORD
            else:
                math_coord = self.OFFSET_COORD
        if isinstance(math_coord, str):
            math_coord = math_coord.format(name=name)
        self.math_coord = math_coord

    # # # GETTING RELEVANT PARAMS # # #
    def get_N_full(self):
        '''return N as if inclusive=(True, True). infer N if needed,
        and ensuring actual N is within acceptable range N_min < N < N_max.

        might return a non-integer; if that occurs, self.grid() will handle it properly.
        '''
        step_param = self.step_param
        inclusive_adjust = (0 if self.inclusive[0] else 1) + (0 if self.inclusive[1] else 1)
        if step_param == 'N':
            result = self.N + inclusive_adjust  # easiest case possible!!
        else:
            span = self.span
            ratio = self.ratio
            if step_param == 'step':
                if span is None:
                    min, max = self.get_lims()
                    span = max - min
                result = 1 + span / self.step
            elif step_param == 'logstep':
                if ratio is None:
                    min, max = self.get_lims()
                    ratio = max / min
                result = 1 + np.log10(ratio) / self.logstep
            else:
                assert False, 'coding error'
            result = result
        actual_N = int(result - inclusive_adjust)
        intN = int(result)
        if intN == result:
            result = intN
        else:  # self._adjust_if_nonint_N later will set inclusive[0]=True or inclusive[1]=True,
            # so here we need to adjust actual_N by adding 1.
            actual_N = actual_N + 1
        if self.N_min is not None and actual_N < self.N_min:
            raise DimensionSizeError(f'N={actual_N} too small! (N_min={self.N_min})')
        if self.N_max is not None and actual_N > self.N_max:
            raise DimensionSizeError(f'N={actual_N} too large! (N_max={self.N_max})')
        return result

    def get_N(self):
        '''return N, appropriate given self.inclusive.
        
        might return a non-integer; if that occurs, self.grid() will handle it properly.
        '''
        N_full = self.get_N_full()
        inclusive_adjust = (0 if self.inclusive[0] else 1) + (0 if self.inclusive[1] else 1)
        return N_full - inclusive_adjust

    def get_lims(self):
        '''return (self.min, self.max), inferring (from span or ratio) if needed.'''
        min = self.min
        max = self.max
        span = self.span
        ratio = self.ratio
        if span is not None:
            if min is None:
                min = max - self.span
            if max is None:
                max = min + self.span
        elif ratio is not None:
            if min is None:
                min = max / self.ratio
            if max is None:
                max = min * self.ratio
        return min, max

    def _adjust_if_nonint_N(self, min, max, N, inclusive):
        '''returns min, max, N, and inclusive; adjusted appropriately if N is not an integer.
        (if N is an integer, returns min, max, N unchanged.)
        (N should be N_full, i.e. the value of N as if inclusive=(True, True); e.g. from self.get_N_full())

        The adjustment depends on self.inclusive.
        if inclusive = (True, True), crash.
        else, use original step (or logstep) while dropping the input value for max or min.
            (if inclusive=(False, True), keep original max, adjust original min;
             else, adjust original max, keep original min.)

        N will always be rounded down to the nearest integer.
        '''
        intN = int(N)
        if N == intN:
            return min, max, intN, inclusive
        elif inclusive == (True, True):
            errmsg = (f'non-integer N={N} incompatible with inclusive=(True, True);\nProbably provided '
                      'step (or logstep) which does not evenly divide span (or ratio) between max and min.\n'
                      'Consider using inclusive=(True,False) or (False,False), to use original step (or logstep),\n'
                      'while dropping the input value for max.')
            raise InputConflictError(errmsg)
        elif inclusive[1]:
            if self.step_param == 'step':
                min = max - (intN-1) * self.step
                inclusive = (True, inclusive[1])  # adjusted min such that final point is at new min; old min still not included.
            elif self.step_param == 'logstep':
                min = max / 10**((intN-1) * self.logstep)
                inclusive = (True, inclusive[1])  # adjusted min such that final point is at new min; old min still not included.
            else:
                raise NotImplementedError(f'[TODO] handle non-integer N when step_param={self.step_param!r}')
        else:
            if self.step_param == 'step':
                max = min + (intN-1) * self.step
                inclusive = (inclusive[0], True)  # adjusted max such that final point is at new max; old max still not included.
            elif self.step_param == 'logstep':
                max = min * 10**((intN-1) * self.logstep)
                inclusive = (inclusive[0], True)  # adjusted max such that final point is at new max; old max still not included.
            else:
                raise NotImplementedError(f'[TODO] handle non-integer N when step_param={self.step_param!r}')
        return min, max, intN, inclusive

    # # # CREATING THE GRID # # #
    grid = alias('__call__')

    def __call__(self):
        '''return grid from self.'''
        if self.array_lims:
            return self.array_grid()
        else:
            return self.scalar_grid()

    def scalar_grid(self):
        '''return grid from scalar min to scalar max, with int(self.get_N()) points.
        fails with InputError if min or max are arrays (i.e. if self.array_lims.)
        '''
        if self.array_lims:
            raise InputError('cannot make scalar grid when min or max are arrays.')
        N = self.get_N_full()
        min, max = self.get_lims()
        inclusive = self.inclusive
        min, max, N, inclusive = self._adjust_if_nonint_N(min, max, N, inclusive)
        name = self.name
        d = self.dim
        if self.logspace:
            grid = np.logspace(np.log10(min), np.log10(max), N)
        else:
            grid = np.linspace(min, max, N)
        if not inclusive[0]:
            grid = grid[1:]
        if not inclusive[1]:
            grid = grid[:-1]
        if self.reverse:
            grid = grid[::-1]
        result = xr.DataArray(grid, dims=[d], coords={name: (d, grid)}, name=name)
        return result

    def math_grid01(self):
        '''return XarrayGrid(0, 1, N=self.get_N()).grid(), maybe removing endpoints (see: inclusive),
        and labeled for helping with array_grid().
        '''
        N = self.get_N()
        math_coord = '_internal_name_' if self.math_coord==False else self.math_coord
        grid01_obj = XarrayGrid(0, 1, N=N, name=math_coord, dim=self.dim, inclusive=self.inclusive)
        result = grid01_obj.scalar_grid()
        if self.math_coord==False:
            result = result.drop_vars('_internal_name_')
        return result

    def array_grid(self):
        '''return grid from array min to array max, with self.get_N() points along grid dimension.
        fails with InputError if min and max are both not arrays (i.e. not self.array_lims)
        '''
        if not self.array_lims:
            raise InputError('cannot make array grid when min and max are not arrays.')
        min, max = self.get_lims()
        dim = self.dim
        # make an evenly spaced grid from 0 to 1 then do array math to expand range to min & max.
        grid01 = self.math_grid01()
        if self.logspace:
            logmin = np.log10(min)
            logmax = np.log10(max)
            result = 10**(logmin + (logmax - logmin) * grid01)
        else:
            result = min + (max - min) * grid01
        result = result.rename(self.name).assign_coords({self.name: result})
        # reverse if needed
        if self.reverse:
            result = result.isel({dim: slice(None, None, -1)})
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        contents = []
        for mname, m in (('min', self.min), ('max', self.max)):
            if isinstance(m, xr.DataArray):
                name = '' if m.name is None else f'name={m.name!r}, '
                m = f'DataArray({name}sizes={dict(m.sizes)})'
            elif callable(m):
                if hasattr(m, '__name__'):
                    m = m.__name__
                else:
                    m = f'(callable of type={type(m).__name__})'
            contents.append(f'{mname}={m}')
        if self.span is not None:
            contents.append(f'span={self.span}')
        contents.append(f'{self.step_param}={getattr(self, self.step_param)}')
        if self.step_param == 'N' and self.logspace:
            contents.append('logspace=True')
        if self.inclusive != (True, True):
            contents.append(f'inclusive={self.inclusive}')
        contents.append(f'dim={self.dim!r}')
        return f'{type(self).__name__}({", ".join(contents)})'

@format_docstring(xarray_grid_docs=XarrayGrid.__doc__)
def xarray_grid(min=None, max=None, N=None, name='grid', *, step=None, span=None, ratio=None,
                 logspace=None, logstep=None, inclusive=(True, True), reverse=False,
                 dim=None, math_coord=False, N_max=UNSET, N_min=UNSET):
    '''create and return a grid, via XarrayGrid(...).grid().
    (If you want to inspect the grid-making process more closely, use XarrayGrid instead)

    XarrayGrid docs copied below for convenience:
    ---------------------------------------------
    {xarray_grid_docs}
    '''
    grid = XarrayGrid(min=min, max=max, N=N, name=name, step=step, span=span, ratio=ratio,
                      logspace=logspace, logstep=logstep, inclusive=inclusive, reverse=reverse,
                      dim=dim, math_coord=math_coord, N_max=N_max, N_min=N_min)
    return grid()


''' --------------------- xarray_angle_grid --------------------- '''

def xarray_angle_grid(min=0, max=360, N=None, name='ang', *, rad_in=False,
                      step=None, span=None,
                      radstep=None, degstep=None, radspan=None, degspan=None,
                      inclusive=(True, False), N_min=UNSET, N_max=UNSET):
    '''create and return a grid of angles; output will always be in [radians].
    (The default is to create a grid from 0 to 360 degrees, including 0 but not 360)

    The main advantage of this method is to make simple 1D angle grids.
    For more precise control, and/or 2D+ grids, use xarray_grid or XarrayGrid instead.

    min: None, number, or callable
        min value. None --> infer from max & span.
        callable f --> infer min [radians] = f(max [radians])
    max: None, number, or callable
        max value. None --> infer from min & span.
        callable f --> infer max [radians] = f(min [radians])
    N: None or number
        number of points in grid. None = "unspecified"
    name: str
        result.name; also result.assign_coords({name:result})
    rad_in: bool, default False.
        whether the inputs are in radians (if True) or degrees (if False).
    step: None or number
        step size between points, in same units as input (depends on rad_in). None = "unspecified"
    radstep: None or number
        step size between points, in radians. None = "unspecified"
    degstep: None or number
        step size between points, in degrees. None = "unspecified"
    span: None or number
        max - min, in same units as input (depends on rad_in). None = "unspecified"
    radspan: None or number
        max - min, in radians. None = "unspecified"
    degspan: None or number
        max - min, in degrees. None = "unspecified"
    inclusive: tuple of 2 bools
        whether to include min and max in the output.
    N_min: UNSET, None or number
        if provided, tells minimum allowed number of points in grid.
        None --> no minimum. UNSET --> use XarrayGrid.N_min default (==2)
    N_max: UNSET, None or number
        if provided, tells maximum allowed number of points in grid.
        None --> no maximum. UNSET --> use XarrayGrid.N_max default (==1e8)
    '''
    steps = dict(step=step, radstep=radstep, degstep=degstep)
    if sum([1 for v in steps.values() if v is not None]) > 1:
        raise InputConflictError(f'cannot provide more than one of: step, radstep, degstep; but got {steps}')
    spans = dict(span=span, radspan=radspan, degspan=degspan)
    if sum([1 for v in spans.values() if v is not None]) > 1:
        raise InputConflictError(f'cannot provide more than one of: span, radspan, degspan; but got {spans}')
    # convert all to radians:
    if not rad_in:
        if (min is not None) and (not callable(min)):
            min = np.deg2rad(min)
        if (max is not None) and (not callable(max)):
            max = np.deg2rad(max)
        if step is not None:
            step = np.deg2rad(step)
        if span is not None:
            span = np.deg2rad(span)
    if step is None:
        step = (None if degstep is None else np.deg2rad(degstep)) if radstep is None else radstep
    if span is None:
        span = (span if degspan is None else np.deg2rad(degspan)) if radspan is None else radspan
    # create & return grid:
    return xarray_grid(min=min, max=max, N=N, name=name, step=step, span=span,
                       inclusive=inclusive, N_min=N_min, N_max=N_max)



"""
File Purpose: loading single-fluid Bifrost quantities related to Equation Of State (EOS).
"""
from ...mhd import MhdEosLoader
from ...tools import alias


''' --------------------- BifrostEosLoader--------------------- '''

class BifrostEosLoader(MhdEosLoader):
    '''single-fluid Bifrost quantities related to Equation of State (EOS): ne, T, P.

    The implementation here assumes tables available at table=self.tabin[var],
        for var='ne', 'T', or 'P', and each having a table.interp(r=r, e=e) method,
        which gives value of var in 'raw' units.
    '''

    # non-NEQ functionality is inherited from MhdEosLoader.
    # NEQ ("nonequilibrium") functionality is implemented here.

    EOS_MODE_OPTIONS = {**MhdEosLoader.EOS_MODE_OPTIONS,
        'neq': '''non-equilibrium ionization for H (possibly also for He too):
            ne and T from hionne and hiontg (from aux). P from table, r, and e.'''
    }

    _default_eos_mode = alias('eos_mode_sim')

    def eos_mode_sim(self):
        '''how simulation handled "Equation of State" related variables (ne, T, P).
        (provides default value for self.eos_mode.)

        'ideal' --> treated as ideal gas: P = n kB T = (gamma - 1) e.
            ne not available.
        'table' --> plugged into EOS lookup tables (see self.tabin)
            plug r and e into tables to get ne, T, P.
        'neq' --> non-equilibrium ionization for H (possibly also for He too):
            ne and T from hionne and hiontg (from aux). P from table, r, and e.
        '''
        if 'tabinputfile' not in self.params:
            return 'ideal'
        elif self.params.get('do_hion', False):
            return 'neq'
        else:
            return 'table'

    # tell super().get_ne to use self('ne_neq') if eos_mode=='neq':
    _EOS_MODE_TO_NE_VAR = {**MhdEosLoader._EOS_MODE_TO_NE_VAR, 'neq': 'ne_neq'}

    # tell super().get_T to use self('T_neq') if eos_mode=='neq':
    _EOS_MODE_TO_T_VAR = {**MhdEosLoader._EOS_MODE_TO_T_VAR, 'neq': 'T_neq'}

    # tell super().get_P to use self('P_fromtable') if eos_mode=='neq':
    _EOS_MODE_TO_P_VAR = {**MhdEosLoader._EOS_MODE_TO_P_VAR, 'neq': 'P_fromtable'}
    # (even in 'neq' mode, P still comes from table, r, and e.)

    @known_var(dims=['snap'])
    def get_ne_neq(self):
        '''electron number density, from 'hionne' in aux.
        hionne in aux is stored in cgs units.
        '''
        ufactor = self.u('n', convert_from='cgs')
        return self.load_maindims_var_across_dims('hionne', u=ufactor, dims=['snap'])

    @known_var(dims=['snap'])
    def get_T_neq(self):
        '''temperature, from 'hiontg' in aux.
        hiontg in aux is stored in [K] units.
        '''
        # note: multifluid T_neq assumes same T for all fluids.
        return self.load_maindims_var_across_dims('hiontg', u='K', dims=['snap'])



"""
File Purpose: test docs_tools.
"""

import PlasmaCalcs as pc

def test_docstring_to_sphinx():
    '''ensures that DocstringInfo(docstring).to_sphinx() succeeds,
    for all modules, classes, functions, and methods in PlasmaCalcs.
    '''
    ll = pc.list_objs(pc)
    for obj in ll:
        doc = obj.__doc__
        if doc is not None:
            di = pc.DocstringInfo(doc)
            di.to_sphinx()  # just making sure it doesn't crash :)



"""
File Purpose: test solving TFBI theory, from eppic inputs
TFBI = Thermal Farley-Buneman Instability.

See e.g.:
    - [Evans+2025](https://doi.org/10.3847/1538-4357/adcd70)
    - [Evans+2023](https://doi.org/10.3847/1538-4357/acc5e5)
    - [tfbi_theory](https://pypi.org/project/tfbi_theory)
    - PlasmaCalcs/examples/tfbi_theory.ipynb
"""
import os

import matplotlib.pyplot as plt
import numpy as np

import PlasmaCalcs as pc

HERE = os.path.dirname(__file__)
INPUT_DECKS = os.path.join(HERE, 'test_tfbi_inputs')
TESTS_DIR = os.path.dirname(HERE)
ARTIFACTS = os.path.join(TESTS_DIR, '_test_artifacts', 'tfbi')
if os.path.exists(ARTIFACTS):  # remove any xarray_io artifacts before running tests.
    import shutil; shutil.rmtree(ARTIFACTS)
os.makedirs(ARTIFACTS)

''' --------------------- helper functions --------------------- '''

def get_ec(eppic_i, u_t=1, kw_units=dict(M=1), **kw_init):
    '''make an eppic calculator with SI units by default'''
    with pc.InDir(INPUT_DECKS):
        ec = pc.EppicCalculator.from_here(eppic_i, u_t=u_t, kw_units=kw_units, **kw_init)
    return ec

def get_ec_2a(**kw_ec):
    '''get_ec('eppic_2a_tinybox.i')'''
    return get_ec('eppic_2a_tinybox.i', **kw_ec)

def savefig_and_close(name, *, bbox_inches='tight', **kw):
    '''savefig to ARTIFACTS/name, then plt.close().'''
    plt.savefig(os.path.join(ARTIFACTS, name), bbox_inches=bbox_inches, **kw)
    plt.close()


''' --------------------- ensure can import tfbi_theory --------------------- '''

def test_import_tfbi_theory():
    '''ensure that tfbi_theory can be imported'''
    import tfbi_theory


''' --------------------- solve TFBI from eppic params --------------------- '''

def test_solve_tfbi_2a():
    '''solve TFBI theory directly from eppic params.'''
    ec = get_ec_2a()
    ec.snap = pc.INPUT_SNAP  # don't look for snapshot data; get vals from input deck.
    solver = ec.tfbi_solver()
    with pc.TimeLimit(20):  # (should only take ~5 seconds. Crash if taking way too long.)
        dsR = solver.solve()
    # make and save plot (auto test just ensures it doesn't crash; can inspect by eye later.)
    dsR.it.growthplot()
    savefig_and_close('tfbi_2a_growthplot.png')
    # compare with known solution:
    #  ("known" meaning: I checked it when originally making test)
    assert np.abs(dsR.it.growth_kmax() - 1353) < 10
    assert np.isclose(np.rad2deg(dsR.it.kang_at_growmax()), 60)


''' --------------------- solve TFBI from "sort of" eppic params --------------------- '''

def test_solve_tfbi_2a_ish():
    '''solve TFBI theory from eppic params but varying some params artificially.'''
    ec = get_ec_2a()
    ec.snap = pc.INPUT_SNAP  # don't look for snapshot data; get vals from input deck.
    tfbivals0 = ec.tfbi_ds()   # other options include: ec('tfbi_inputs'), ec('tfbi_all')
    Etest = pc.xr1d([4, 6, 10, 20], 'E_test')
    tfbivals = tfbivals0.assign({'E_un0_perpmod_B': Etest})
    cc = pc.InstabilityCalculator(tfbivals)
    cc_tfbivals = cc.tfbi_ds()
    # ensure recomputing eqperp_ldebye using new E.
    #  (tfbivals0['eqperp_ldebye'] doesn't vary with E_test.)
    assert 'E_test' in cc_tfbivals['eqperp_ldebye'].dims
    # solve TFBI theory, use higher kres
    #   also use larger mod range so that more klines can be checked by eye.
    solver = cc.tfbi_solver(kres='mid', mod=dict(min=10**-1, max=10**3.5))
    with pc.TimeLimit(20):  # (should only take ~5 seconds. Crash if taking way too long.)
        dsR = solver.solve()
    # make and save plot (auto test just ensures it doesn't crash; can inspect by eye later.)
    dsR.it.growthplots(share_vlims='all')  # share_vlims to use same colorbar for all plots
    savefig_and_close('tfbi_2a_growthplots_Etest.png')
    # use InstabilityCalculator interface to compute some values, to ensure it works
    cc2 = pc.InstabilityCalculator(dsR)
    # compare with known solution:
    #  ("known" meaning: I checked it when originally making test)
    growthmax = cc2('growth_kmax')  # max growth rate (max taken across k) at each point
    assert growthmax.isel(E_test=0).item() < 0  # no growth at E==4
    assert np.allclose(growthmax.isel(E_test=slice(1,None)), [810,  2290,  3560], rtol=1e-2, atol=0)
    kang = cc2('rad2deg_kang_at_growmax').isel(E_test=slice(1,None))
    assert np.allclose(kang, [57., 63., 72.], rtol=1e-4, atol=2)
    kmod = cc2('kmod_at_growmax').isel(E_test=slice(1,None))
    assert np.allclose(kmod, [7.28, 5.09, 3.56], rtol=1e-2, atol=0)



"""
File Purpose: Dimension values & lists for Copapic
"""
import numpy as np

from ...dimensions import Fluid, FluidList
from ...errors import InputError, InputMissingError, FluidKeyError
from ...tools import (
    alias, value_from_aliases,
    DictlikeFromKeysAndGetitem,
    format_docstring,
    UNSET,
)


''' --------------------- CopapicDist and CopapicDistList --------------------- '''

class CopapicDist(DictlikeFromKeysAndGetitem, Fluid):
    '''info about a single distribution.
    see cls.from_dict() to create an instance from a dict of values from an Copapic.json file.
    units are "raw". Numerical values equal those from the Copapic.json file.

    m: real number
        mass of a single "real" particle in the distribution
    q: real number
        charge of a single "real" particle in the distribution
    n0: real number
        background "physical" number density of this distribution of particles.
    params: None or dict
        additional parameters.
    name: None or str
        name of this distribution.
        Optional, and probably not found in Copapic.json files.
        Provide to improve human-readability.
    '''
    precision = 2  # precision when converting values to human-readable string

    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'name', 'm', 'q', 'n0', 'params'}

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, m, q, n0, name, i, *, params=None):
        params = dict() if params is None else params
        super().__init__(name=name, i=int(i), m=m, q=q)
        self.n0 = n0
        self.N = int(i)
        self.params = params

    @classmethod
    def from_dict_and_N(cls, d, N):
        '''return CopapicDist from dict and N.
        d: dict
            result of copapic_io_tools.read_copapic_json_file()[N].
        N: None, str or int
            distribution number. if int, will be converted to str.
        '''
        N = str(N)
        params = {key : value for key, value in d.items()}
        return cls(m=params['mass'], q=params['charge'], n0=params['n0'], name=params['name'], i = N, params=params)

    @classmethod
    def dists_from_dict(cls, d):
        '''return dict of distributions, given a dict of {N: dict of distribution values}.'''
        return {N: cls.from_dict_and_N(d, N) for N, d in d.items()}

    # # # DISPLAY # # #
    def __repr__(self):
        fmt = f'{{:.{self.precision}e}}'
        contents = []
        if self.N is not None:
            contents.append(f'N={self.N}')
        if self.name is not None:
            contents.append(f'name={self.name!r}')
        contents.extend([f'{a_}={fmt.format(getattr(self, a_))}' for a_ in ('m', 'q', 'n0')])
        content_str = ', '.join(contents)
        return f'{type(self).__name__}({content_str})'

    # # # GETTING VALUES / ITERATING # # #
    SPECIAL = ('m', 'q', 'n0', 'N', 'name')

    @format_docstring(special=SPECIAL)
    def __getitem__(self, key):
        '''usually return self.params[key]; if key is special return self.key instead.
        special keys are determined by self.SPECIAL; default {special}.
            else, return self.params[key].
        '''
        if key in self.SPECIAL:
            return getattr(self, key)
        try:
            return self.params[key]
        except KeyError:
            errmsg = f'key {key!r} not in {type(self).__name__}.params, and not special {self.SPECIAL}'
            raise KeyError(errmsg) from None

    def keys(self):
        '''return tuple of keys which are special or in self.params'''
        return tuple(key for key in (*self.SPECIAL, *self.params.keys()))

    # # # QUASINEUTRAL? # # #
    def is_hybrid(self):
        '''return whether this distribution is "hybrid", i.e. fluid model instead of PIC.
        as a proxy, guessing hybrid iff self.get('method', 0) != 0.
        '''
        return self.get('method', 0) != 0

    # # # GETTING SOME PARAMETERS # # #
    def get_n0(self):
        '''returns background (mean) density of this distribution. Equivalent: self['n0']'''
        return self['n0']

    def get_v0(self, x=None):
        '''returns original (t=0) velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).
        '''
        if x is None:
            return (self.get_v0('x'), self.get_v0('y'), self.get_v0('z'))
        else:
            key = f'v{x}0d'
            return self[key]

    def get_vth0(self, x=None):
        '''returns original (t=0) thermal velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).

        if x is Ellipsis (i.e., x=...), assert all three components are the same,
            and return the one value which they all equal.
        '''
        if x is None:
            return (self.get_vth0('x'), self.get_vth0('y'), self.get_vth0('z'))
        elif x is ...:
            vx, vy, vz = self.get_vth0('x'), self.get_vth0('y'), self.get_vth0('z')
            assert vx == vy == vz, f'Expected all equal, but got: vth0x={vx}, vth0y={vy}, vth0z={vz}'
            return vx
        else:
            key = f'v{x}thd'
            return self[key]

    def get_velocity_coords(self, units):
        '''
        Returns a dictionary of vdist coordinates (vx, vy, vz) for this CopapicDist.
        For example, "vx": np.linspace(self["pvxmin"], self["pvxmax"], num=self["pnvx"]).
        '''
        coords = {
            "vx": np.linspace(self["pvxmin"], self["pvxmax"], num=self["pnvx"]) * units,
            "vy": np.linspace(self["pvymin"], self["pvymax"], num=self["pnvy"]) * units,
            "vz": np.linspace(self["pvzmin"], self["pvzmax"], num=self["pnvz"]) * units
        }
        return coords

    def get_npd(self, *, default=UNSET):
        '''return number of PIC particles for this distribution in each processor.
        If that is not possible (due to using alternate option for npd), raise FluidKeyError
        Alternate options include: nptotd, npcelld, nptotcelld;
            those options tell npd info but scaled by number of cells and/or processors.
            number of cells & processors are global parameters not known by CopapicDist,
            so CopapicDist cannot calculate npd from those options.

        default: UNSET or any value
            return default instead of error when npd not found, if default provided (not UNSET).
        '''
        npd = self.get('npd', None)
        if npd is None:
            if default is not UNSET:
                return default
            # else, raise FluidKeyError; error message depends on which alternative option in self.keys().
            keys = self.keys()
            errmsg = "'npd' not provided, and CopapicDist cannot determine 'npd' from "
            if 'nptotd' in keys:
                errmsg += "'nptotd', because CopapicDist doesn't know number of processors."
            elif 'npcelld' in keys:
                errmsg += "'npcelld', because CopapicDist doesn't know number of cells."
            elif 'nptotcelld' in keys:
                errmsg += "'nptotcelld', because CopapicDist doesn't know number of cells and processors."
            else:
                errmsg = "No indication of 'npd' was provided to CopapicDist."
            raise FluidKeyError(errmsg)
        # else:
        return npd


class CopapicDistList(FluidList):
    '''list of CopapicDist objects'''
    value_type = CopapicDist


''' --------------------- CopapicNeutral and CopapicNeutralList --------------------- '''

class CopapicNeutral(Fluid):
    '''info about the neutral fluids, in Copapic.
    [raw] units system; numerical values match input deck.

    m: number
        mass of neutral fluid particle.
    name: str, default 'neutral'
        name of the neutral fluid.
    i: int, default 0
        Copapic dist that this neutral fluid is associated with.
    vth: None or number
        thermal velocity of the neutral fluid. None --> unknown.
    v0: None or number
        drift velocity of the neutral fluid. None --> unknown.
    '''
    precision = 2  # precision when converting values to human-readable string

    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'m', 'name', 'i', 'vth'}

    def __init__(self, m, *, name='neutral', i=0, vth=None, v0=None):
        self.vth = vth
        self.v0 = v0
        super().__init__(name, m=m, q=0, i=i)

    # # # GETTING SOME PARAMETERS # # #
    def get_vth(self, x=None):
        '''returns thermal velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).
        (Note that for neutrals, vth is the same for all 3 components.)
        if self.vth is None, raise InputMissingError.
        '''
        if self.vth is None:
            raise InputMissingError(f'cannot get_vth when {type(self).__name__}.vth is None')
        if x is None:
            return (self.vth, self.vth, self.vth)
        else:
            return self.vth
    
    def get_v0(self, x=None):
        '''returns drift velocity, as a tuple of (vx, vy, xz) values,
        or the x component if x provided (x should be a string if provided).
        if self.v0 is None, raise InputMissingError.
        '''
        if self.v0 is None:
            raise InputMissingError(f'cannot get_v0 when {type(self).__name__}.v0 is None')
        if x is None:
            return tuple(self.v0)
        else:
            if x == 'x':
                return self.v0[0]
            elif x == 'y':
                return self.v0[1]
            elif x == 'z':
                return self.v0[2]
            else:
                raise InputError(f'get_v0: x={x} not in {{x,y,z}}')
            
    # # # ALIASES # # #
    get_vth0 = alias('get_vth',
            doc='''alias to self.get_vth. Provides interface consistency with CopapicDist.get_vth0.
            note: neutral vth is the same for all 3 components, and doesn't vary with time.''')
    get_v0d = alias('get_v0',
            doc='''alias to self.get_v0. Provides interface consistency with CopapicDist.get_v0d.
            note: neutral v0 doesn't vary with time.''')

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(m={self.m:.{self.precision}e}, vth={self.vth[0]:.{self.precision}e}, v0=({self.v0[0]:.{self.precision}e}, {self.v0[1]:.{self.precision}e}, {self.v0[2]:.{self.precision}e}))'

    # # # COMPARISON # # # 
    # sometimes it's convenient to have an ordering for the values...
    # and, let's make sure CopapicNeutral always compares as "less than" any CopapicDist.
    # [TODO] better code encapsulation (some of this is repeated from DimensionValue class)?
    def __lt__(self, other):
        if isinstance(other, CopapicDist): return True
        elif isinstance(other, type(self)): return (self.i, self.s) < (other.i, other.s)
        else: return NotImplemented
    def __le__(self, other):
        if isinstance(other, CopapicDist): return True
        if isinstance(other, type(self)): return (self.i, self.s) <= (other.i, other.s)
        else: return NotImplemented
    def __gt__(self, other):
        if isinstance(other, CopapicDist): return False
        if isinstance(other, type(self)): return (self.i, self.s) > (other.i, other.s)
        else: return NotImplemented
    def __ge__(self, other):
        if isinstance(other, CopapicDist): return False
        if isinstance(other, type(self)): return (self.i, self.s) >= (other.i, other.s)
        else: return NotImplemented


class CopapicNeutralList(FluidList):
    '''list of CopapicNeutral objects.'''
    value_type = CopapicNeutral

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)



"""
File Purpose: Component, ComponentList, ComponentDimension, ComponentHaver
"""
import xarray as xr

from .dimension_tools import (
    DimensionValue, DimensionValueList, string_int_lookup,
    Dimension, DimensionHaver,
)
from ..errors import ComponentKeyError, ComponentValueError, InputError
from ..tools import is_integer

''' --------------------- Component & ComponentList --------------------- '''

class Component(DimensionValue):
    '''a single vector component coordinate (e.g. x, y, or z).
    Knows how to be converted to str and int. e.g. 'y' or 1.

    s: string or None
        str(self) --> s. if None, cannot convert to str.
    i: int or None
        int(self) --> i. if None, cannot convert to int.

    Note: not related to "coordinates", e.g. location in space.
    '''
    def __init__(self, s=None, i=None):
        super().__init__(s, i)


class ComponentList(DimensionValueList):
    '''list of vector component coordinates.'''
    _dimension_key_error = ComponentKeyError
    value_type = Component

    @classmethod
    def from_strings(cls, strings):
        '''return ComponentList from iterable of strings. (i will be determined automatically.)'''
        return cls(Component(s, i) for i, s in enumerate(strings))

    @classmethod
    def from_array(cls, array):
        '''return ComponentList from 0D or 1D array.
        array values should be Component objects, strings, or ints.
            Component objects --> use as-is
            strings --> if 'x', 'y', or 'z', use XYZ.get(string).
                        Else, use Component(s, i=i) for i in range(len(array))
            ints --> if 0, 1, or 2, use XYZ.get(int).
                     Else, use Component(i=i) for i in array
        '''
        # [TODO] use super() instead of repeating code from DimensionValueList.from_array?
        if array.ndim == 0:
            values = [array.item()]
        elif array.ndim == 1:
            values = array.values
        else:
            errmsg = f'{cls.__name__}.from_array expects array.ndim=0 or 1; got ndim={array.ndim}'
            raise DimensionalityError(errmsg)
        result = []
        for i, v in enumerate(values):
            if isinstance(v, cls.value_type):
                result.append(v)
            elif (isinstance(v, str) or isinstance(v, int)) and v in cls.value_type.XYZ:
                result.append(cls.value_type.XYZ.get(v))
            elif isinstance(v, str):
                result.append(cls.value_type(v, i=i))
            elif is_integer(v):
                result.append(cls.value_type(i=v))
            else:
                errmsg = (f'{cls.__name__}.from_array got unexpected value type at index {i}.\n'
                            f'Expected {cls.value_type.__name__}, str, or int; got value={v!r}.')
                raise InputError(errmsg)
        return cls(result)
        

XYZ = ComponentList.from_strings('xyz')
X, Y, Z = XYZ
YZ_FROM_X = string_int_lookup({X:(Y,Z), Y:(Z,X), Z:(X,Y)})  # right-handed coord system x,y,z given x.

Component.YZ_FROM_X = YZ_FROM_X  # add to Component class for convenience.
Component.XYZ = XYZ  # add to Component class for convenience.

XHAT = xr.DataArray([1,0,0], coords={'component': XYZ})
YHAT = xr.DataArray([0,1,0], coords={'component': XYZ})
ZHAT = xr.DataArray([0,0,1], coords={'component': XYZ})


''' --------------------- ComponentDimension, ComponentHaver --------------------- '''

class ComponentDimension(Dimension, name='component', plural='components',
                     value_error_type=ComponentValueError, key_error_type=ComponentKeyError):
    '''component dimension, representing current value AND list of all possible values.
    Also has various helpful methods for working with this Dimension.
    '''
    pass  # behavior inherited from Dimension.


@ComponentDimension.setup_haver
class ComponentHaver(DimensionHaver, dimension='component', dim_plural='components'):
    '''class which "has" a ComponentDimension. (ComponentDimension instance will be at self.component_dim)
    self.component stores the current vector component (possibly multiple). If None, use self.components instead.
    self.components stores "all possible vector components" for the ComponentHaver.
    Additionally, has various helpful methods for working with the ComponentDimension,
        e.g. current_n_component, iter_components, take_component.
        See ComponentDimension.setup_haver for details.

    components defaults to XYZ (==ComponentList.from_strings('xyz'))
    '''
    def __init__(self, *, component=None, components=XYZ, **kw):
        super().__init__(**kw)
        if components is not None: self.components = components
        self.component = component



"""
File Purpose: CopapicCalculator
"""

import numpy as np
import xarray as xr

from .copapic_bases import CopapicBasesLoader
from .copapic_direct_loader import CopapicDirectLoader
from .copapic_input_deck import CopapicInputDeck
from .copapic_io_tools import read_copapic_snaps_info
from ...dimensions import SnapList, INPUT_SNAP
from ...errors import DimensionError, FluidValueError
from ...plasma_calculator import MultifluidPlasmaCalculator
from ...tools import (
    alias, alias_child,
    UNSET,
)
from ...defaults import DEFAULTS

class CopapicCalculator(CopapicBasesLoader, CopapicDirectLoader, MultifluidPlasmaCalculator):
    '''PlasmaCalculator for Copapic outputs.
    To initialize from an copapic.json file, use CopapicCalculator.from_here(fname).
    
    Use CopapicCalculator.help() for more help.

    input_deck: CopapicInputDeck
        the CopapicInputDeck which is being used by this calculator.
    u: None or UnitsManager
        if provided, set self.u = u, and ignore u_l, u_t, u_n, and ne_si.
    u_l, u_t, u_n, ne_si: None or value
        must provide one of these values in order to fully determine the units.
        u_l: length [si] = u_l * length [raw]
        u_t: time [si] = u_t * time [raw]
        u_n: number density [si] = u_n * number density [raw]
        ne_si: electron number density [si]. (ne [raw] will be loaded from input_deck)
    kw_units: dict
        any additional values used to determine the units.
        can also provide u_l, u_t, u_n, or ne_si, here, instead of as direct kwarg.
        See help(UnitsManagerPIC.from_pic) for more details.
    quasineutral: None or bool
        if provided, set self.quasineutral.
        otherwise, infer it from self.input_deck.is_quasineutral().
        in quasineutral mode, distribution 0 is assumed to be electrons,
            and different routines are used to calculate its density & other parameters.
    _more_init_checks: bool, default False
        if True, perform additional checks during initialization.
        False by default for slight efficiency improvement,
            and to avoid possibility of init checks failing but object still being usable.

    other kwargs are passed to super().__init__.
    '''
    # # # DEFAULTS # # #
    _quasineutral = False

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, input_deck, *, 
                 u=None, u_l=None, u_t=None, u_n=None, ne_si=None, kw_units=dict(),
                 quasineutral=None,
                 _more_init_checks=False,
                 **kw_super):
        self.input_deck = input_deck
        kwu = kw_units.copy()
        self.init_units_manager(u=u, u_l=kwu.pop('u_l', u_l), u_t=kwu.pop('u_t', u_t),
                                u_n=kwu.pop('u_n', u_n), ne_si=kwu.pop('ne_si', ne_si), **kwu)
        self.init_maindims()
        self.init_fluids()
        self.init_snaps()
        # kw_super.setdefault('jfluid', 0)
        if quasineutral is None: quasineutral = self.input_deck.is_quasineutral()
        super().__init__(input_deck, quasineutral=quasineutral, **kw_super)
        if _more_init_checks:
            self._check_files_readable()

    @classmethod
    def from_here(cls, filename, **kw):
        '''create CopapicCalculator from input deck file found here (at filename).
        dist_names: None or dict
            {N: name for distribution N} for any number of distributions in filename.
            E.g.: {0: 'e', 1: 'H+', 2: 'C+'}
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return cls(CopapicInputDeck.from_file(filename), **kw)

    # # # INFO # # #
    dirname = alias_child('input_deck', 'dirname')
    filename = alias_child('input_deck', 'filename')

    # # # UNITS # # #
    def init_units_manager(self, *, u=None, u_l=None, u_t=None, u_n=None, ne_si=None, **kw_get_units_manager):
        '''set units manager self.u, based on input_deck.
        must provide one of these values in order to fully determine the units.
            u: UnitsManager object. if provided, set self.u = u.
            u_l: length [si] = u_l * length [raw]
            u_t: time [si] = u_t * time [raw]
            u_n: number density [si] = u_n * number density [raw]
            ne_si: electron number density [si]. (ne [raw] will be loaded from input_deck)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if u is None:
            u = self.input_deck.get_units_manager(u_l=u_l, u_t=u_t, u_n=u_n, ne_si=ne_si, **kw_get_units_manager)
        self.u = u

    # # # DIMENSIONS SETUP -- MAINDIMS # # #
    def init_maindims(self):
        '''set self.maindims based on input_deck.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.maindims = self.input_deck.maindims()

    def get_maindims_coords(self, units=None):
        '''return dict of {'x': xcoords, 'y': ycoords}. Possibly also 'z': zcoords.
        units: None or str
            result will be in this unit system. if None, use self.coords_units (default: self.units).
        coords will also be sliced according to self.slices.
        '''
        result = self.input_deck.get_space_coords()
        u_l = self.u('l', self.coords_units_explicit if units is None else units)
        result = {key: u_l * value for key, value in result.items()}
        return self._apply_maindims_slices_to_dict(result)

    get_space_coords = alias('get_maindims_coords')

    # # # DIMENSIONS SETUP -- OTHERS # # #
    def init_snaps(self):
        '''set self.snaps based on input_deck AND reading files.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        dt = self.input_deck["dt"]
        dir_ = self.input_deck.dirname
        output_dir = self.input_deck.output_dir
        names, times = read_copapic_snaps_info(output_dir, dt, dir_)
        self.snaps = SnapList.from_lists(s=names, t=times)

    def init_fluids(self):
        '''set self.fluids and self.jfluids based on input_deck.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.fluids = self.input_deck.dists
        self.jfluids = self.input_deck.neutrals

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of strings to use inside self.__repr__'''
        contents = super()._repr_contents()
        contents = [f'input_deck={self.input_deck!r}'] + list(contents)
        return contents


"""
File Purpose: Muram Calculator
"""
import os

import numpy as np

from .muram_bases import MuramBasesLoader
from .muram_direct_loader import MuramDirectLoader
from .muram_eos_loader import MuramEosLoader
from .muram_snaps import MuramSnapList
from ...mhd import  MhdCalculator, MhdUnitsManager
from ...tools import alias


class MuramCalculator(MuramEosLoader, MuramBasesLoader,
                      MuramDirectLoader,
                      MhdCalculator):
    '''PlasmaCalculator for Muram outputs.
    
    dir: str
        directory where files are located. Stored as self.dirname = os.path.abspath(dir).
    units: 'si' or 'raw'
        units system for outputs of self. 'raw' for same as muram data, 'si' for SI units.
        'cgs' for cgs, but note cgs electromagnetic quantities are not yet implemented.
        Implementation here assumes Muram outputs are always in cgs, and that mu0_raw = 1;
            i.e. MhdUnitsManager.from_mhd_cgs(ucgs_l=1, ucgs_t=1, ucgs_r=1, mu0_raw=1)
    '''

    def __init__(self, *, dir=os.curdir, units='si', **kw_super):
        self.dirname = os.path.abspath(dir)
        self.init_snaps()
        self.u = MhdUnitsManager.from_mhd_cgs(ucgs_l=1, ucgs_t=1, ucgs_r=1, mu0_raw=1)
        self.units = units
        super().__init__(**kw_super)

    # # # DIMENSIONS SETUP -- SNAPS # # #
    def init_snaps(self):
        '''set self.snaps based on snap files in self.dirname.'''
        self.snaps = MuramSnapList.from_here(dir=self.dirname)

    # # # PARAMS # # #
    @property
    def params(self):
        '''return the global params shared by all snaps.
        Equivalent to self.snaps.params_global(recalc=False).
        '''
        return self.snaps.params_global(recalc=False)

    # # # DIMENSIONS SETUP -- MAINDIMS # # #
    @property
    def maindims(self):
        '''maindims of MURAM data, i.e. ('x', 'y', 'z').
        (currently only 3D MURAM data reading is implemented.)
        see MainDimensionsHaver for more maindims implementation details.
        '''
        return ('x', 'y', 'z')

    get_space_coords = alias('get_maindims_coords')

    def get_maindims_coords(self):
        '''return dict of {'x': xcoords, 'y': ycoords, 'z': zcoords}.
        Units will be whatever is implied by self.coords_units system (default: self.units)
        coords will be sliced according to self.slices, if relevant.
        '''
        maindims = self.maindims
        u_l = self.u('l', self.coords_units_explicit)
        dx = {x: self.params[f'd{x}'] * u_l for x in maindims}
        Nx = {x: self.params[f'N{x}'] for x in maindims}
        result = {x: np.arange(Nx[x]) * dx[x] for x in maindims}
        result = self._apply_maindims_slices_to_dict(result)
        return result

    # # # INPUT TABLES (self.tabin) # # #
    @property
    def tabinputfile(self):
        '''abspath to tabinputfile (used when creating default self.tabin).
        return os.path.join(self.dirname, 'tabparam.in')
        '''
        return os.path.join(self.dirname, 'tabparam.in')



"""
File Purpose: ComparisonLoader.
"""
# [TODO] add tests for the patterns here, similar to basic arithmetic tests,
#   to ensure order of operations behaves as expected.

from ..quantity_loader import QuantityLoader
from ...tools import (
    UNSET, simple_property,
    xarray_at_max_of, xarray_at_min_of,
)

class ComparisonLoader(QuantityLoader):
    '''comparison operations, e.g. ==, !=, >, <, >=, <=.
    Also, some tools for boolean arrays: 'not_{var}', '{var0}_or_{var1}', '{var0}_and_{var1}'.
    Also, 'where': self('var_where_condition') --> self('var').where(self('condition'))
    Also, 'at_max_of' and 'at_min_of', for convenient argmaxing.
    '''
    # Caution: the order of these functions matters!
    # Patterns are checked in the order that functions were defined.
    # E.g., define 'or' before 'and' --> 'A_or_B_and_C' becomes 'A' or 'B_and_C'.
    # The order of definitions here matches Python standard order:
    #   not_A_or_B <--> (not_A)_or_B
    #   not_A_and_B <--> (not_A)_and_B
    #   A_or_B_and_C <--> A_or_(B_and_C)
    #   A_and_B_or_C <--> (A_and_B)_or_C
    # Additionally, logic (or, and, not) defined before comparisons (e.g. ==, >, <),
    #   ensures that, e.g., 'A==B_and_C' is parsed as '(A==B)_and_C'.
    #   (NOT parsed as 'A==(B_and_C)'.)

    @known_pattern(r'not_(.+)', deps=[0])  # 'not_{var}'
    def get_logical_not(self, var, *, _match=None):
        '''logical not. not var. Equivalent: ~self(var)'''
        var, = _match.groups()
        return ~self(var)

    @known_pattern(r'(.+)_or_(.+)', deps=[0, 1])  # '{var0}_or_{var1}'
    def get_logical_or(self, var, *, _match=None):
        '''logical or. var0 or var1. Equivalent: self(var0) | self(var1)'''
        var0, var1 = _match.groups()
        return self(var0) | self(var1)

    @known_pattern(r'(.+)_and_(.+)', deps=[0, 1])  # '{var0}_and_{var1}'
    def get_logical_and(self, var, *, _match=None):
        '''logical and. var0 and var1. Equivalent: self(var0) & self(var1)'''
        var0, var1 = _match.groups()
        return self(var0) & self(var1)

    # # # VALUE COMPARISONS # # #
    @known_pattern(r'(.+)[=][=](.+)', deps=[0,1])  # '{var0}=={var1}'
    def get_compare_equals(self, var, *, _match=None):
        '''self('A==B') --> boolean array: self('A') == self('B')'''
        var0, var1 = _match.groups()
        return self(var0) == self(var1)

    @known_pattern(r'(.+)[!][=](.+)', deps=[0,1])  # '{var0}!={var1}'
    def get_compare_not_equals(self, var, *, _match=None):
        '''self('A!=B') --> boolean array: self('A') != self('B')'''
        var0, var1 = _match.groups()
        return self(var0) != self(var1)

    # note - must define >= before >, otherwise e.g. 'a>=b' --> 'a' > '=b'.
    @known_pattern(r'(.+)[>][=](.+)', deps=[0,1])  # '{var0}>={var1}'
    def get_compare_greater_than_or_equal(self, var, *, _match=None):
        '''self('A>=B') --> boolean array: self('A') >= self('B')'''
        var0, var1 = _match.groups()
        return self(var0) >= self(var1)

    @known_pattern(r'(.+)[>](.+)', deps=[0,1])  # '{var0}>{var1}'
    def get_compare_greater_than(self, var, *, _match=None):
        '''self('A>B') --> boolean array: self('A') > self('B')'''
        var0, var1 = _match.groups()
        return self(var0) > self(var1)

    # note - must define <= before <, otherwise e.g. 'a<=b' --> 'a' < '=b'.
    @known_pattern(r'(.+)[<][=](.+)', deps=[0,1])  # '{var0}<={var1}'
    def get_compare_less_than_or_equal(self, var, *, _match=None):
        '''self('A<=B') --> boolean array: self('A') <= self('B')'''
        var0, var1 = _match.groups()
        return self(var0) <= self(var1)

    @known_pattern(r'(.+)[<](.+)', deps=[0,1])  # '{var0}<{var1}'
    def get_compare_less_than(self, var, *, _match=None):
        '''self('A<B') --> boolean array: self('A') < self('B')'''
        var0, var1 = _match.groups()
        return self(var0) < self(var1)

    # # # VAR_WHERE_CONDITION # # #
    cls_behavior_attrs.register('drop', default=UNSET)
    drop = simple_property('_drop', default=UNSET,
        doc='''value of 'drop' kwarg for any self('{var}_where_{condition}') calls.
        True --> drops points where condition is False. (See xarray.DataArray.where for details)
        False --> use nan where condition is False.
        default: UNSET --> True if self('condition') has ndim==1, else False.
            (easy when ndim==1 to drop nans, because condition is a 1D list of points.
             hard when ndim>=2. E.g. if mask (x,y)=(0,0) but not (0,1) and (1,0), can't drop x=0 nor y=0...)''')

    @known_pattern(r'(.+)_where_(.+)', deps=[0,1])  # '{var}_where_{condition}'
    def get_var_where_condition(self, var, *, _match=None):
        '''var_where_condition --> self(var).where(self(condition)).
        for 'drop' kwarg, in where(..., drop=...), use drop=self.drop.
        '''
        var, condition = _match.groups()
        condval = self(condition)
        drop = self.drop
        if drop is UNSET: drop = (condval.ndim == 1)
        return self(var).where(condval, drop=drop)

    @known_pattern(r'where_(.+)_(.+)', deps=[0,1])  # where_{condition}_{var}
    def get_where_condition_var(self, var, *, _match=None):
        '''where_condition_var --> self(var).where(self(condition)).
        Note: if var contains any underscores, must use parenthesis (like 'where_condition_(var)').
            The alias, self('var_where_condition'), does not have such a restriction.
        for 'drop' kwarg, in where(..., drop=...), use drop=self.drop_in_where.
        '''
        condition, var = _match.groups()
        return self(f'{var}_where_{condition}')

    # # # AT_MAX/MIN_OF # # #
    @known_pattern(r'(.+)_at_max_of_(.+)', deps=[0,1])  # '{var}_at_max_of_{ref}'
    def get_var_at_max_of_ref(self, var, *, _match=None):
        '''var_at_max_of_ref --> self(var) at argmax of self(ref),
        taking argmax across all dims in self(ref).
        For more precise control, consider directly using xarray_at_max_of.
        '''
        here, ref = _match.groups()
        return xarray_at_max_of(self(here), self(ref))

    @known_pattern(r'(.+)_at_min_of_(.+)', deps=[0,1])  # '{var}_at_min_of_{ref}'
    def get_var_at_min_of_ref(self, var, *, _match=None):
        '''var_at_min_of_ref --> self(var) at argmin of self(ref),
        taking argmin across all dims in self(ref).
        For more precise control, consider directly using xarray_at_min_of.
        '''
        here, ref = _match.groups()
        return xarray_at_min_of(self(here), self(ref))



"""
Package Purpose: misc plotting tools.
"""
# module references, in case other packages need to access internal modules.
# provide module references for all modules which might be overwritten by same-named object.
# E.g. from .colorbar import colorbar --> "colobar" here will be the function not the module.
from . import colorbar as _colorbar_module

from .colorbar import (
    make_cax, find_mappable, colorbar,
    Colorbar,
    _paramdocs_colorbar,
)
from .colors import (
    get_cmap,
    ColormapExtremes, BaseColormap, Colormap, cmap,
    CMAPS,
)
from .currently_active import (
    current_figure_exists, current_figure_or_None,
    current_axes_exists, current_axes_or_None, current_axes_has_data,
    current_image_exists, current_image_or_None,
    maintaining_current_plt, maintaining_current_figure, maintaining_current_axes, maintaining_current_image,
    using_current_plt, using_current_figure, using_current_axes, using_current_image,
    MaintainingCurrentPlt, UsingCurrentPlt,
)
from .lims import (
    calculate_vlims, update_vlims, calculate_lims_from_margin,
    get_data_interval, get_lims_with_margin,
    plt_zoom, plt_zoomx, plt_zoomy, set_margins_if_provided,
    set_lims_if_provided,
)
from .plot_dims import (
    PlotDimsMixin,
    infer_movie_dim, infer_xy_dims, infer_xyt_dims,
)
from .plot_tricks import (
    ax_outline, fig_outline, ax_remove_ticks, ax_remove_ticklabels,
    get_colorbar_axes,
    get_min_n_ticks, set_min_n_ticks,
    use_simple_log_tick_locator, SimpleLogTickLocator, use_simple_ticks_renamer,
    ax_aspect,
    plt_transformer,
    PLOT_LOCATION_NAMES, plot_locations, plot_note,
)



"""
File Purpose: EppicHybridCalculator, for outputs from eppic hybrid simulations.
"""

import xarray as xr

from .eppic_calculator import EppicCalculator
from ...dimensions import IONS
from ...errors import FluidValueError, FormulaMissingError

class EppicHybridCalculator(EppicCalculator):
    '''
    EppicHybridCalculator is a subclass of EppicCalculator that is designed to
    handle hybrid simulations in EPPIC. It provides methods for loading and
    processing data specific to hybrid simulations.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hybrid = True

    @known_var(load_across_dims=['fluid'])
    def get_distribution_type(self):
        '''distribution type: DataArray of strings: 'electron', 'ion', or 'neutral'
        (useful internally for handling electrons differently,
            since electrons are fluid while ions are PIC particles.)
        '''
        if self.fluid.is_electron():
            return xr.DataArray('electron')
        elif self.fluid.is_ion():
            return xr.DataArray('ion')
        elif self.fluid.is_neutral():
            return xr.DataArray('neutral')
        else:
            raise FluidValueError(f'expected electron, ion, or neutral fluid; got {self.fluid}')

    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_deltafrac_n(self, *, distribution_type):
        '''normalized density perturbation. deltafrac_n = (n - n0) / n0.
            For hybrid simulations, electron density is the sum of the ion densities  
        '''
        if distribution_type == 'electron':
            n0 = self.get_n0()
            n = self('sum_fluids_n', fluid=IONS)
            return (n - n0) / n0
        else:
            return super().get_deltafrac_n()

    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_flux(self, *, distribution_type):
        '''flux. (for non-electrons: directly from Eppic. for electrons: not yet implemented)
        (Eventually, can determine electron flux via solving electron momentum equation for u_e,
            neglecting the du_e/dt term. Result similar to u_drift, but include P_e contribution.)
        '''
        if distribution_type == 'electron':
            raise FormulaMissingError('get_flux for electrons not yet implemented')
        else:
            return super().get_flux()
    
    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_n0(self, *, distribution_type):
        '''background density. (directly from Eppic.)
        For hybrid simulations, electron density is the sum of the ion densities  
        note: as with all other quantities, this will be output in [self.units] unit system;
            numerically equal to eppic.i value if using 'raw' units.
        '''
        if distribution_type == 'electron':
            return self('sum_fluids_n0', fluid=IONS)
        else:
            return super().get_n0()
        
    @known_var(partition_across_dim=('fluid', 'distribution_type'), aliases=['temp', 'temperature'])
    def get_T(self, *, distribution_type):
        ''' temperature in Kelvin.

        For electrons, just loads 'temperature' from snapshot.
        For ions, equivalent to rmscomps_Ta.
            (Use self.T_indim_only=True if you want to ignore Ta_z in 2D sim)
        '''
        # [TODO] actually, should probably define get_Ta_or_Tajoule for electrons,
        #    since that is the "base" var in super().
        #    e.g. I think Ta for electrons could just return 'temperature' (without varying across component);
        #    that might make it just work, but it might also take some debugging to get it right.
        #    - SE (2025/05/07)
        if distribution_type == 'electron':
            result = self.load_maindims_var_across_dims('temperature', dims=['snap']) / self.u('kB')
            return result
        else:
            return super().get_T()
    
    @known_var(partition_across_dim=('fluid', 'distribution_type'))
    def get_T_box(self, *, distribution_type):
        '''temperature of the entire simulation box, as if full of particles,
        and observed by something that could not resolve the individual cells.
        Equivalent: rmscomps(Ta_from_moment2).

        Ignores Ta components for directions in which the box has no extent.
        '''
        if distribution_type == 'electron':
            return self('mean_T')  # Te is isotropic for hybrid eppic --> T_box is just mean_T.
        else:
            return super().get_T_box()
